## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, rietbroek@utwente.nl

import numpy as np
import xarray as xr
from xinv.core.attrs import rhs_attrs,N_attrs,ltpl_attrs,sigma0_attrs,nobs_attrs,npara_attrs,find_xinv_coords, xunk_coords_attrs,aux_coords_attrs,x0_attrs

from xinv.neq.build import build_normal as neq_build_normal

class FwdOpbase:
    def __init__(self,obs_dim=None,unknown_dim=None,cache=False,jacobname="jacobian",**bindargs):
        self._obsdim=obs_dim
        self._unkdim=unknown_dim
        self._unkdim_t=unknown_dim+"_"
        self._daobs=None
        self._cache_jacobian=cache
        self._jacob=None
        self._jacobname=jacobname
        #to be forwarded to the jacobian implementation together with additional arguments
        self._bindargs=bindargs

    @property
    def unkdim(self):
        """Return the unknown dimension"""
        return self._unkdim

    @property
    def obs_dim(self):
        """Return the observation dimension"""
        return self._obsdim
    
    @property
    def jacobname(self):
        """Return the name of the Jacobian variable"""
        return self._jacobname

    def jacobian(self,**kwargs): 
        """Create the Jacobian of the forward operator"""
        
        #possibly quickly return the cached Jacobian
        if self._jacob is not None and self._cache_jacobian:
            return self._jacob

        # if daobs is None and self._jacob is None:
            # raise ValueError("Requesting the Jacobian without arguments requires caching abilities of the forward operator")
        # elif daobs is None:
            # return self._jacob

        jacob=self._jacobian_impl(**self._bindargs,**kwargs)
        if type(jacob) == xr.DataArray:
            jacob.name=self._jacobname
            jacob=jacob.to_dataset()
        #add xunk_coord attributes to possible coordinate
        if self._unkdim in jacob.coords:
            jacob[self._unkdim].attrs.update(xunk_coords_attrs(state='linked'))
        

        if self._cache_jacobian:
            self._jacob=jacob
        return jacob

    def build_normal(self,daobs,ecov=1):
        #convenience function
        #just forward the call to the implementation
        return neq_build_normal(self,daobs,ecov)

    
    def __call__(self,inpara,**kwargs):
        """Apply the forward operator"""
        return self.jacobian(**kwargs)[self._jacobname]@inpara



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import numpy as np
import xarray as xr
from xinv.core.attrs import find_neq_components,x0_attrs
from xinv.core.tools import find_ilocs
from xinv.linalg.inplace import dsymm_inplace
from xinv.core.logging import xinvlogger

def set_x0(dsneqin:xr.Dataset, daapri:xr.DataArray, apriori_is_delta=False, inplace=False):
    """ Change apriori values in a normal equation system """
    if inplace:  
        dsneq=dsneqin.copy()
    else:
        #note: xarray deep copy does really always do a deep copy of the underlying data (e.g. numpy ndarrays)
        dsneq=dsneqin.copy(deep=True)
    
    
    
    #note prefix io_ denotes used for in and output, i_ is used for input only
    i_N,io_rhs,io_x0,io_ltpl,_,_,_=find_neq_components(dsneq)
    
    unkdim=dsneq.xi.unknown_dim()
    
    if io_x0 is None:
        #create a new  x0 entry
        dsneq['x0']=xr.zeros_like(io_rhs.reset_index(unkdim))
        dsneq.x0.attrs=x0_attrs()
        io_x0=dsneq.x0
    
    if not inplace:
        #todo: handle non-ndarrays here (e.g. dask,or sparse arrays)
        #copy data explicitly (xarray deep copy does not always seem to work as expected)
        #breakpoint() 
        io_rhs.data=io_rhs.data.copy(order='F')
        #link back to output neq
        dsneq[io_rhs.name]=io_rhs
        
        #io_x0.data=io_x0.data.copy(order='F')
        #dsneq[io_x0.name]=io_x0

        #io_ltpl.data=io_ltpl.data.copy()

    #import ipdb;ipdb.set_trace()



    #find the indices of the apriori values in the unknown vector
    idxapri=find_ilocs(dsneq,unkdim,daapri.coords[unkdim])
    
    deltax0=xr.zeros_like(io_rhs.reset_index(unkdim))
    aslc={unkdim:idxapri}
    if apriori_is_delta:
        # The apriori input is used as an increment to the existing x0
        deltax0[aslc]=daapri
        io_x0+=deltax0
    else:
        # The input is the new target apriori value and the deltax needed is computed from the existing x0
        deltax0[aslc]=daapri-io_x0[aslc]
        io_x0[aslc]=daapri
    
    
    # (1) first update to ltpl
    io_ltpl[()]-=deltax0.dot(io_rhs,dim=unkdim)
    
    # (2) update right hand side with symmetric matrix multiplication
    dsymm_inplace(A=i_N,B=deltax0,C=io_rhs,alpha=-1.0,beta=1.0)
    #ipdb.set_trace()
    # (3) update ltpl again
    io_ltpl[()]-=deltax0.dot(io_rhs,dim=unkdim)
    

    return dsneq



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl


# fixtures shared among test cases

import pytest
import numpy as np
import xarray as xr
import os
import xinv 

# we fix the seed for the noise generator below to make it noisy but reproducible
rg=np.random.default_rng(12789)

@pytest.fixture(params=['C','F'])
def noisypoly(request):
    """
    Generate a noisy dataset from a set of polynomial functions
    """

    npoly=3
    noise_std=0.01
    x0=0
    delta_x=2
    naux=10 #number of auxiliary datasets
    x_axis=np.arange(-2,3,0.05)
    x_axis_rel=(x_axis-x0)/delta_x
    order=request.param[0]
    polyobs=np.zeros([naux,len(x_axis)],order=order)
    polytrue=np.zeros([naux,npoly+1])
    for i in range(naux):
        polytrue[i,:]=[j*i for j in range(npoly+1)]
        polyobs[i,:]=np.polyval(polytrue[i,::-1],x_axis_rel)
    
    #add some normal noise
    polyobs+= rg.normal(0,noise_std,polyobs.shape)
    
    # create a naming of the auxdims
    auxcoord=[f"aux_{i}" for i in range(naux)]
    dspoly=xr.Dataset({"polyobs":xr.DataArray(polyobs,dims=("naux","x")),"polytrue":xr.DataArray(polytrue.T,dims=("poly","naux"))},coords={"x":x_axis,"naux":auxcoord,"poly":np.arange(npoly+1)},attrs=dict(noise_std=noise_std,x0=x0,delta_x=delta_x))

    return dspoly


@pytest.fixture
def neqbase(request):
    """
    Load a normal equation system from file 
    """
    if request.param=='simple':
        #simple normal equation system
        neqfile1=os.path.join(os.path.dirname(__file__),f'testdata/neqpoly_simple.nc')
        dsneq=xr.load_dataset(neqfile1)
    else:
        neqfile1=os.path.join(os.path.dirname(__file__),f'testdata/neqpoly.nc')
        dsneq=xr.load_dataset(neqfile1).xi.reindex_groups()
    
    return dsneq



#!/usr/bin/env python3
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv import *
from xinv.fwd.polynomial import Polynomial
import os

from fixtures import noisypoly

neqpolyf=os.path.join(os.path.dirname(__file__),f'testdata/neqpoly_simple.nc')



def test_polyneqs(noisypoly):
    """
    Test the polynomial forward operator, building of a normal equation system, and solving step
    Parameters
    ----------
    noisypoly : xr.Dataset containing the noisy polynomial observations and the true polynomial coefficients

    """
    #initialize the forward operator
    x0=noisypoly.attrs['x0']
    deltax=noisypoly.attrs['delta_x']
    npoly=noisypoly.sizes['poly']-1
    polyfwd=Polynomial(n=npoly,poly_x='x',x0=x0,delta_x=deltax,cache=True)
    
    #build the normal equation system
    std_noise=0.5
    dsneq=noisypoly.polyobs.xi.build_normal(polyfwd,ecov=std_noise*std_noise) 
    if not os.path.exists(neqpolyf):
        #save the normal equation system to a file (for use in other tests)
        dsneq.to_netcdf(neqpolyf)

    dssol=dsneq.xi.solve()
    
    #1. check whether the posteriori sigma is decently estimated
    prenoise=noisypoly.attrs['noise_std']
    #3 sigma test
    assert np.allclose(dssol.sigma0,prenoise,rtol=0.25)

    tol=3*np.sqrt(np.diag(dssol.COV))*prenoise
    tol=tol.max()

    assert np.allclose(dssol.solution,noisypoly.polytrue,atol=tol)        

    #4. propagate the solution to the observations

    fwdobs=polyfwd(dssol.solution)
    # test for containment within 6 sigma 
    np.allclose(noisypoly.polyobs,fwdobs.T,atol=6*prenoise) 




# This file is part of the xinv software which is licensed
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

# distutils: language = c++
# cython: profile=False

import numpy as np
cimport numpy as np
import xarray as xr
import os
from libc.stdlib cimport strtol,strtod
from xinv.core.logging import xinvlogger
from xinv.core.attrs import N_attrs, get_xunk_size_coname
import cython

#A bit of a hack to directly allow access to the data of a Python unicode str  object
cdef extern from *:
    void *PyUnicode_DATA(object unicode)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def read_symmat_fast(fileobj,dsout,blockname):
    """
        Reads a triangular matrix from a SINEX block and returns a symmetric version (Cython version)
        This version is optimized and uses the following tricks
        1. Direct access through a character pointer to the data of a unicode object (line of the files)
        2. Use of Cython memoryviews to avoid boiunds checking of the numpy arrays
        3. Use of Cython cdef variables to avoid type checking and optimize loop
        4. Use of stdlib strtol, and strod to parse the lines
    Parameters
    ----------
    fileobj : 
        io buffer to read lines from
        
    dsout : xarray.Dataset
        xarray.Dataset to augment the matrix data to
        
        
    blockname : str
        name of the SINEX block. should be one of:
        SOLUTION/NORMAL_EQUATION_MATRIX U
        SOLUTION/NORMAL_EQUATION_MATRIX L
        

    Returns
    -------
    an updated xarray.Dataset holding the new matrix in a new variable 'N'

    """
    if not blockname.startswith('SOLUTION/NORMAL_EQUATION_MATRIX'):
        raise RuntimeError(f"Wrong block {blockname}?")
    cdef int nest
    nest,xunk_co=get_xunk_size_coname(dsout)
    xunkdim=dsout[xunk_co].dims[0]
    mat=np.zeros([nest,nest],order='C')
    
    cdef double [:,:] cmat =mat
    cdef int irow
    cdef int icol
    cdef int ndat
    # cdef const unsigned char[:] ucline
    cdef char* cline
    cdef str line
    for line in fileobj:
        #get the char pointer to the buffer in the unicode object
        cline= <char*> PyUnicode_DATA(line)
        # cline = <const char*> &line[0]
        #cline = line[0]
        if cline[0] == '-':
            #end of block encountered
            break
        elif cline[0] == '*':
            #comment
            continue

        # sscanf(cline, " %5li %5li", &irow, &icol)
        irow=strtol(cline,&cline,10)

        icol=strtol(cline,&cline,10)
        irow-=1 #note zero indexing
        icol-=1

        ndat=nest-icol if icol>nest-3 else 3
        
        if ndat == 1:
            # #read one element
            cmat[irow,icol]=strtod(cline,&cline)
        elif ndat == 2:
            # #read 2 elements
            cmat[irow,icol]=strtod(cline,&cline)
            cmat[irow,icol+1]=strtod(cline,&cline)
            
        else:
            # #read 3 elements
            cmat[irow,icol]=strtod(cline,&cline)
            cmat[irow,icol+1]=strtod(cline,&cline)
            cmat[irow,icol+2]=strtod(cline,&cline)
    
    #mirror the upper triangle in the lower part
    mat=np.triu(mat,k=1).T+mat

    # if "nm_" not in dsout.indexes:
        # #add the transposed index
        # mi_=SHindexBase.mi_toggle(dsout.indexes['nm'])
        # dsout=dsout.sh.set_nmindex(mi_,'_')

    dsout['N']=([xunkdim,f"{xunkdim}_"],mat)
    dsout.N.attrs.update(N_attrs())
    return dsout




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl, 2025 Kiana Karimi, s.karimi@utwente.nl

import xarray as xr 
from xinv.fwd import FwdOpbase
from xinv.core.attrs import xunk_coords_attrs
class MatrixfwdOp(FwdOpbase):
    def __init__(self,dajac:xr.DataArray):
        obs_dim=dajac.dims[0]
        unknown_dim=dajac.dims[1]
        super().__init__(obs_dim=obs_dim,unknown_dim=unknown_dim,cache=True)
        #sets the cached Jacobian matrix, so _jacobian_impl never gets called 
        self._jacob=dajac.to_dataset(name=self._jacobname)
        if self._unkdim in dajac.coords:
            self._jacob[self._unkdim].attrs.update(xunk_coords_attrs(state='linked'))








            



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl
import numpy as np
import xarray as xr
from xinv.core.attrs import find_component, get_xunk_size_coname, group_id_attrs, group_seq_attrs,find_xinv_coords, xunk_coords_attrs,xinv_tp,xinv_st
import pandas as pd
from xinv.core.logging import xinvlogger

def find_group_coords(dsneq,grpdim=None,assoc_coords=None):
    """
    Find the group id and sequence coordinates in a dataset
    Parameters
    ----------
    dsneq : xr.Dataset
        Contains Normal equations system elements or a solution thereof
    grpdim: str
        name of the dimension which matches the group coordinates
    Returns
        
    -------
    (group_id_co,group_seq_co) : xr.DataArray
        The group id and seq coordinate or (None,None) if not found
    
    """
    group_id_co=None
    group_seq_co=None
    
    #try to find heuristically by naming
    if grpdim is not None:
        for k,var in dsneq.variables.items():
            if len(var.dims) > 0 and var.dims[0] == grpdim:
                if k.endswith('grp_id'):
                    group_id_co=dsneq[k]
                elif k.endswith('grp_seq'):
                    group_seq_co=dsneq[k]
    else:
        try:
            group_id_co=find_component(dsneq,xinv_tp.grp_id_co)
            group_seq_co=find_component(dsneq,xinv_tp.grp_seq_co)
        except KeyError:
            pass
        
    #try to find the associated coordinates found in the group_id
    group_asso_co={}
    if group_id_co is not None:
        for asso_co_name in np.unique(group_id_co):

            if asso_co_name in dsneq:
                group_asso_co[asso_co_name]=dsneq[asso_co_name]
            elif assoc_coords is not None and asso_co_name in assoc_coords:
                group_asso_co[asso_co_name]=assoc_coords[asso_co_name]
            else:
                xinvlogger.warning(f"Missing associated group id coordinate: {asso_co_name}, consider adding from original source")
                group_asso_co[asso_co_name]=None


        
    return group_id_co,group_seq_co,group_asso_co

def build_group_coord(data,dim='xinv_unk',group_id_name="xinv_grp_id",group_seq_name="xinv_grp_seq"):
   
    if type(data) == xr.DataArray:
        data=data.data
    if type(data[0]) == tuple:
        mi=pd.MultiIndex.from_tuples(data, names=[group_id_name,group_seq_name])
    else:
        mi=pd.MultiIndex.from_arrays(data, names=[group_id_name,group_seq_name])

    grpcoord=xr.Coordinates.from_pandas_multiindex(mi,dim=dim)
    #add the proper xinv attributes
    grpcoord[dim].attrs.update(xunk_coords_attrs(state=xinv_st.linked))
    #note we need to re-add the group id and seq coordinate attributes
    grpcoord[group_id_name].attrs.update(group_id_attrs(state=xinv_st.linked))
    grpcoord[group_seq_name].attrs.update(group_seq_attrs(state=xinv_st.linked))
    return grpcoord

def expand_as_group(dsin,group_dim,group_id_dim="xinv_grp_id",group_seq_dim="xinv_grp_seq",stack_dim=None):
    """
    Expand a DataArray or Dataset along a group dimension, adding a group_id_dim and group_seq_dim to the dataset
    Parameters
    ----------
    dsin : xr.DataArray or xr.Dataset
        The input data to expand
        
    group_dim : str
        The dimension in the original data to expand
        
    group_id_dim : str, optional
        The name of the new dimension to hold the group id
        
    group_seq_dim : str, optional
        The name of the new dimension to hold the sequence number within the group
        
    stack_dim : str, optional
        The name of the new dimension to stack the group_id_dim and group_seq_dim
        

    Returns
    -------
    xr.Dataset with the expanded coordinates
        

    """
    dsout=dsin.expand_dims(dim=group_id_dim).rename({group_dim:group_seq_dim}).assign_coords({group_id_dim:(group_id_dim,[group_dim]),group_seq_dim:(group_seq_dim,np.arange(len(dsin[group_dim])))})
    if type(dsout) == xr.DataArray:
        dsout=dsout.to_dataset()
    # dsout=dsout.rename({group_dim:group_seq_dim})

    #For retrieval purposes later: copy the original parameters back to the old dimension
    dsout=dsout.assign_coords({group_dim:dsin[group_dim]})
    
    # add xinv attributes to mark the original coordinates as unlinked
    dsout[group_dim].attrs.update(xunk_coords_attrs(state="unlinked"))    
    dsout[group_id_dim].attrs.update(group_id_attrs(state="linked"))
    dsout[group_seq_dim].attrs.update(group_seq_attrs(state="linked"))

    if stack_dim is not None:
        dsout=dsout.stack({stack_dim:[group_id_dim,group_seq_dim]})
        dsout[stack_dim].attrs.update(xunk_coords_attrs(state="linked"))
    
    return dsout



def get_group(dsneq,groupname):
    """
    Retrieve a parameter group from a Dataset which contains a Normal equation system or solution thereof
    Parameters
    ----------
    dsneq : xr.Dataset
        Contains Normal equations system elements or a solution thereof
        
    groupname : str
        Groupname to retrieve
    
    returns: xr.Dataset
        The subset of the input which is valid for the groupname

    """
    xinvcoords=find_xinv_coords(dsneq)
    try:
        if type(dsneq.indexes[groupname]) == pd.MultiIndex:
            groupismindex=True
        else:
            groupismindex=False
    except: 
        groupismindex=False

    #create an index vector of the data asssociated with the groupname
    
    #find the group id and sequence dimensions
    group_id_name=find_component(dsneq,xinv_tp.grp_id_co).name
    group_seq_name=find_component(dsneq,xinv_tp.grp_seq_co).name
    
    
    grpidx=dsneq[group_id_name] == groupname
    
    try:
        unkdim=find_component(dsneq,xinv_tp.rhs).dims[0]
    except KeyError:
        unkdim=find_component(dsneq,xinv_tp.solest).dims[0]
    unkdim_=f"{unkdim}_"

    try:
        try:
            find_component(dsneq,xinv_tp.N)
        except KeyError:
            #ok so try find a covariance matrix
            find_component(dsneq,xinv_tp.COV)
        #extract the relevant subsections
        dsout=dsneq[{unkdim:grpidx.data,unkdim_:grpidx.data}] 
        rename={unkdim:groupname,unkdim_:groupname+"_"}
        hasmatrix=True
    except KeyError:
        #neither N or COV was found
        dsout=dsneq[{unkdim:grpidx.data}] 
        rename={unkdim:groupname}
        hasmatrix=False

    
    groupcoord=dsout[groupname][dsout[group_seq_name]]
    
    #multindex and all levels must be dropped
    dropvars=[nm for nm in dsout.get_index(unkdim).names]
    dropvars.append(unkdim)
    if groupismindex:
        #also add all levels of the group to drop
        dropvars.extend([nm for nm in dsout.get_index(groupname).names])

    #also drop the groupname varibale (will be readded later)
    dropvars.append(groupname)
    dsout=dsout.drop_vars(dropvars).rename(rename)
    #extract the relevant original coordinate components
    if groupismindex:
        groupcoord=pd.MultiIndex.from_tuples(groupcoord.data,names=dsneq.get_index(groupname).names)
        groupcoord=xr.Coordinates.from_pandas_multiindex(groupcoord,dim=groupname)
        dsout=dsout.assign_coords(groupcoord)
    else:

        dsout=dsout.assign_coords({groupname:(groupname,groupcoord.data)})
    #possibly rebuild the multindex if the original group was a multindex
    #add xinv attributes
    dsout[groupname].attrs.update(xunk_coords_attrs(state=xinv_st.linked))

    return dsout

def reindex_groups(dsneq,group_dim=None,assoc_coords=None):
    """
    Rebuilds the groups,sequences into a multiIndex of a dataset containing a Normal equation system or solution thereof (e.g. read from a file)
    Parameters
    ----------
    dsneq : xr.Dataset
        Contains Normal equations system elements or a solution thereof
    group_dim:str
        name of a specific group_dimension to look for group coordinates
    assoc_coords: dict
        dictionary with auxiliary source coordinates which can be copied when they are associated with the group_id's
    returns: xr.Dataset
        An xarray.Dataset with a valid multindex

    """
    
    

    
#    try:
 #       NorCOV=find_component(dsneq,xinv_tp.N)
 #   except KeyError:
 #       NorCOV=find_component(dsneq,xinv_tp.COV)
    
 #   unkdim=NorCOV.dims[0]
    if group_dim is None:
        try:
            _,group_dim=get_xunk_size_coname(dsneq)
        except StopIteration:
            #try in a different way
                
            try:
                NorCOV=find_component(dsneq,xinv_tp.N)
            except KeyError:
                NorCOV=find_component(dsneq,xinv_tp.COV)
    
            group_dim=NorCOV.dims[0]
 
    #try to find a group id and sequence dimensions
    group_id_co,group_seq_co,asso_co=find_group_coords(dsneq,group_dim,assoc_coords)
        
    if group_id_co is None or group_seq_co is None:
        raise RuntimeError("Group id and sequence coordinates can not be found from xinv attributes")
    

    #recreate the multindex based on the group and seq id
    grp_co=build_group_coord([dsneq[group_id_co.name].data,dsneq[group_seq_co.name].data],dim=group_dim,group_id_name=group_id_co.name,group_seq_name=group_seq_co.name)
    
    #possibly augment with associated coordinates

    #import pdb; pdb.set_trace()
    grp_co.update(asso_co)

    dsneq=dsneq.drop_vars([group_id_co.name,group_seq_co.name]).assign_coords(grp_co)

    return dsneq


def rename_groups(dsneq,grpmap):
    #try to find a group id and sequence dimensions
    group_id_co,group_seq_co,_=find_group_coords(dsneq)
    if group_id_co is None or group_seq_co is None:
        raise RuntimeError("Group id and sequence coordinates can not be found from xinv attributes, no consistent renaming possible")
    #first remap the coordinate names themselves
    dsneq=dsneq.rename(grpmap)
    #figure out the linked unknown parameter dimensions

    xinvcoords=find_xinv_coords(dsneq,include=xinv_tp.unk_co,state=xinv_st.linked)
    unk_name=next(iter(xinvcoords.keys()))
    dsneq=dsneq.reset_index(unk_name) 
    for grp,grpnew in grpmap.items():
        dsneq[group_id_co.name]=xr.where(dsneq[group_id_co.name] == grp,grpnew,dsneq[group_id_co.name])
        #update attributes
        dsneq[group_id_co.name].attrs.update(group_id_attrs(state=xinv_st.linked))
    #rebuild the index


    return reindex_groups(dsneq)


def split_as_groups(dsneq,group_ids:xr.DataArray,stack_dim=None):
    """
    """
    
    unksz,unkdim=get_xunk_size_coname(dsneq)

    if unksz != len(group_ids):
        raise ValueError(f"Size of group_ids {len(group_ids)} does not match the size of the unknown dimension {unksz} in the dataset")

    #make new groupcoordinates
    grpseqtrack={grp:0 for grp in np.unique(group_ids.data)}
    splitcoords={grp:[] for grp in grpseqtrack.keys()}
    grpdata=[]
    for idx,group_id in enumerate(group_ids.data):
        iseq=grpseqtrack[group_id]
        grpdata.append((group_id,iseq))  # sequence number is always 0 for now
        #copy original coordinate value
        splitcoords[group_id].append(dsneq[unkdim].data[idx])
        grpseqtrack[group_id] += 1
    
    if stack_dim is None:
        stack_dim= f"{unkdim}_spl"
    groupco= build_group_coord(grpdata,dim=stack_dim)
    
    #setup new coordinates
    attrs=xunk_coords_attrs(state=xinv_st.unlinked)
    newcoords={grp:(grp,co,attrs) for grp,co in splitcoords.items()}
    
    dsneq=dsneq.drop_vars([unkdim]).rename({unkdim:stack_dim,f"{unkdim}_":f"{stack_dim}_"}).assign_coords(newcoords).assign_coords(groupco)

    # breakpoint()

    return dsneq



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import xarray as xr
from xinv.core.attrs import find_component, get_xunk_size_coname,xinv_tp,islower,find_neq_components,unlink,find_xinv_coords,get_type
from xinv.core.grouping import split_as_groups,get_group,build_group_coord
from xinv.neq.neq import zeros as neqzeros
from xinv.core.tools import find_overlap_coords,find_ilocs
from xinv.fwd.matrixfwd import MatrixfwdOp
import numpy as np

from scipy.linalg.blas import dtrmm,dsyr2k
from scipy.linalg.lapack import dgesv

def transform(dsneq:xr.Dataset,fwdoperator,apriori_strategy="ignore",**kwargs):
    """Transform a normal equation system using a forward operator

    Parameters
    ----------
    dsneq : xr.Dataset
        The normal equation system to be transformed
    fwdoperator : xinv.fwd.FwdOperator or xarray.DataArray
        The forward operator to transform the normal equation system. This can also be a xarray.DataArray holding the Design matrix
    apriori_strategy : str, optional
        How to treat the transformation of the apriori solution estimate, currently only "ignore" is allowed (sets the output to zero without changing the ltpl)
    
    **kwargs :
        Additional arguments passed to the fwdoperator.jacobian call
    """
   



    if type(fwdoperator) == xr.DataArray:
        # Convenience service: convert the design matrix to a linear forward operator
        fwdoperator=MatrixfwdOp(fwdoperator)
    
    if apriori_strategy not in ["ignore"]:
        raise NotImplementedError(f"apriori strategy {apriori_strategy} not implemented, only 'ignore' is currently allowed")

    #input unknown size and dimension name
    i_unksz,i_unkdim=get_xunk_size_coname(dsneq)

    if fwdoperator.obs_dim != i_unkdim:
        raise ValueError(f"fwdoperator row dimension name {fwdoperator.obs_dim} must match the unknown dimension {i_unkdim}")

    #retrieve the forward operator Jacobian (note: may not be linear, and may need the current apriori solution input to generate the Jacobian)

    #pointers to the input normal equation system
    i_N,i_rhs,i_x0,i_ltpl,i_sigma0,i_nobs,i_npara=find_neq_components(dsneq)
    
    if i_x0 is None:
        i_x0=xr.zeros_like(i_rhs)

    jac=fwdoperator.jacobian(daobs=i_x0,**kwargs)#.reset_index(i_unkdim)
    
    #find unique and overlapping coordinates over the transform dimension
    jauniq,common,neuniq=find_overlap_coords(jac[i_unkdim],dsneq[i_unkdim])
    if len(jauniq) > 0:
        raise ValueError(f"Forward operator has unknown coordinate values {jauniq} which are not found in the input normal equation system {dsneq[i_unkdim]}")
    
    #needed so the rest of the operations using the jacobian does not get messed up with potential name-conflicting indices

    jac=jac.reset_index(i_unkdim)
    #index vector of the input to transformed parameters
    idxtrans=find_ilocs(dsneq,i_unkdim,common)
    idxnotrans=None 
    if len(neuniq) == 0:
        partial=False
        #full transform, all NEQ unknowns are transformed 
        o_unkdim=fwdoperator.unkdim
        outcoords={o_unkdim:jac[o_unkdim]}

    else:
        # this is more complex as some parameters will be removed/introduced, but some remain in the new syste
        partial=True
        o_unkdim='trans_composite'
        o_trgp_id_dim='trans_grp_id'
        o_trgp_seq_dim='trans_grp_seq'
        o_tdim=fwdoperator.unkdim
        
        groups=xr.full_like(dsneq[i_unkdim],i_unkdim,dtype=object)
        groups[idxtrans]='trans'
        
        # i_split_dim=f"{i_unkdim}_split"
        #divide the input system coordinates in two groups 


        # dsneq=split_as_groups(dsneq,groups,stack_dim=i_split_dim)
        
        #new output coordinates (put the untransformed parameters first)
        groupoutco=[(i_unkdim,i) for i in range(len(neuniq))]
        groupoutco.extend([(o_tdim,i) for i in range(jac.sizes[o_tdim])])
        outcoords={o_unkdim:build_group_coord(groupoutco,dim=o_unkdim,group_id_name=o_trgp_id_dim,group_seq_name=o_trgp_seq_dim)}
        #add auxiliary coordinates from input system
        # for k,v in dsneq.coords.items():
            # if k not in ['trans',i_split_dim,'xinv_grp_id', 'xinv_grp_seq']:
                # outcoords[k]=v

        #add output coordinates from transformed system (but make sure to unlink them to the unknown dimension)
        outcoords[o_tdim]=jac[o_tdim]
        unlink(outcoords[o_tdim])

        #also add the untransformed coordinates 
        idxnotrans=find_ilocs(dsneq,i_unkdim,neuniq)
        outcoords[i_unkdim]=dsneq[i_unkdim].isel({i_unkdim:idxnotrans})
        unlink(outcoords[i_unkdim])

   #possibly add coordinates from the Jaconian which are associated with a group coordinate
    

    #add auxiliary coordinates from the input rhs
    for k,v in i_rhs.coords.items():
        if k != i_unkdim:
            if v.dims[0] == i_unkdim:
                
                xinvtype=get_type(v,raiseError=False)
                if xinvtype is not None and (xinvtype == xinv_tp.grp_id_co or xinvtype == xinv_tp.grp_seq_co):
                #don't add coordinated which  are associated with the group coordinates

                    continue
                
                #prevent copying input unknown coordinates
                v=v.drop_vars([i_unkdim])
                if idxnotrans is not None:
                    #Only export the part which refers to the still existent non-transformed part
                    v=v[idxnotrans]
            
            if k in v:
                v=v.drop_vars[k]
            outcoords[k]=v
            #make sure to unlink the coordinates
            unlink(outcoords[k])
    #Allocate the new output normal equations system
    #setup dimensions of the output system
    o_rhsdims=[o_unkdim]
    
    o_rhsdims.extend([dname for dname in i_rhs.dims if dname != i_unkdim])
    
    i_lower=islower(i_N)
    o_dsneq= neqzeros(rhsdims=o_rhsdims,coords=outcoords,lower=i_lower)
    #pointers to the new normal equation system
    o_N,o_rhs,o_x0,o_ltpl,o_sigma0,o_nobs,o_npara=find_neq_components(o_dsneq)

    #setup slicing
    if partial:
        o_tsz=jac.sizes[o_tdim]
        o_unksz=o_rhs.sizes[o_unkdim]
        #slice representing the untransformed part of the unknowns
        uslice=slice(0,o_unksz-o_tsz)
        #slice representing the transformed part of the unknowns
        tslice=slice(o_unksz-o_tsz,None)
    else:
        o_unksz=o_rhs.sizes[o_unkdim]
        #no untransformed part
        uslice=None
        #transformed part is a complete slice
        tslice=slice(None)
        
    #transformed part (valid for both partial and non-partial transforms
    #right hand side
    # add the transformed part of the input rhs to the output rhs (note: inplace operation in o_rhs)
    o_rhs[{o_unkdim:tslice}]=xr.dot(jac.jacobian,i_rhs.isel({i_unkdim:idxtrans}),dim=i_unkdim).data

    #transform normal matrix

        
    #Ntt ndarray will hold the transformed part of the normal matrix
        
    o_Ntt=o_N[{o_unkdim:tslice,o_unkdim+'_':tslice}].data
        
    #Decompose the input Normal matrix  as U^t + U 
    
    # we can apply the following trick: decompose the input matrix i_Ntt in an upper triangular matrix U so
    # i_Ntt = U + U^T, U is the upper triangle of i_Ntt but with the diagonal divided by 2
    
    # The transformation can be written as
    # o_Ntt = A^T (i_Ntt) A  = A^T (U + U^T)  A =  A^T ( U A )  + (A^T U^T) A
    # when UA is written as X we can decompose the entire transform as 
    # (1) extract the triangular matrix U = triu(N) and divide the diagonal by 2
    # (2) compute the transformed matrix as X = U A
    # (3) apply a dsyr2k operation as  Nnew = A^T X + X^T A
        
    #get the index vector to sort N according to the forward operator obs_dim
    if i_lower:
        #extract and mirror lower triangular matrix (mirroring is needed because a resorting may move elements from upper to lower triangle)
        U=np.tril(i_N.data, k=-1) + np.tril(i_N.data, k=-1).T
    else:
        #extract and mirror upper triangular matrix
        U=np.triu(i_N.data, k=1) + np.triu(i_N.data, k=1).T
        
    np.fill_diagonal(U, i_N.data.diagonal()/2)
        
    #resort 
    # idx=find_ilocs(dsneq,i_unkdim,jac[i_unkdim].data)
    # U=U[np.ix_(idx,idx)]
    
    # breakpoint()
    if partial:
        #in case of a partial transform we also need get the cross sectional matrix
        # index of parameters to keep and transform
        idxkeep=find_ilocs(dsneq,i_unkdim,neuniq)
        i_Ntu=U[np.ix_(idxtrans,idxkeep)]

    U=U[np.ix_(idxtrans,idxtrans)]
        
    # retrieve X by a triangular matrix multiplication
    # subroutine dtrmm 	( 	character 	side,
            # character 	uplo,
            # character 	transa,
            # character 	diag,
            # integer 	m,
            # integer 	n,
            # double precision 	alpha,
            # double precision, dimension(lda,*) 	a,
            # integer 	lda,
            # double precision, dimension(ldb,*) 	b,
            # integer 	ldb )
    X=dtrmm(1.0,U,jac.jacobian.data)
       
    # now apply the dsyr2k operation to compute the transformed normal matrixa

        # dsyr2k(alpha, a, b[, beta, c, trans, lower, overwrite_c]) = <fortran function dsyr2k>
    trans=1
    #TODO: avoid memory copy of the output subslice matrix by calling blas from cython (thereby avoiding f2py which makes intermediate copies when passing subarrays)
    o_Ntt[()]=dsyr2k(1.0,jac.jacobian.data,X,trans=trans,lower=i_lower)


    if partial:
        #we need to process the cross terms and copy the untransformed part

        
        #copy the untransformed part of the input normal matrix
        o_N[{o_unkdim:uslice,o_unkdim+'_':uslice}]=i_N.data[np.ix_(idxkeep,idxkeep)]
        
        #copy right hand side
        o_rhs[{o_unkdim:uslice}]=i_rhs[{i_unkdim:idxkeep}].data

        #copy apriori values
        o_x0[{o_unkdim:uslice}]=i_x0[{i_unkdim:idxkeep}].data

        if i_lower:
            #fill lower cross sectional nromal matrix part
            o_N[{o_unkdim:uslice,o_unkdim+"_":tslice}]=i_Ntu.T@jac.jacobian.data
        else:
            o_N[{o_unkdim:tslice,o_unkdim+"_":uslice}]=jac.jacobian.T.data@i_Ntu 




    # copy ltpl's fromt he old system
    o_ltpl[()]=i_ltpl
    o_sigma0[()]=i_sigma0
    o_nobs[()]=i_nobs
    o_npara[()]=i_npara
    #Note we may need to adjust the amount of parameters because some may have been removed or newly introduced
    
    o_npara[()]+=o_unksz-i_unksz 
    return o_dsneq




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl


from xinv.core.attrs import find_xinv_unk_coord,find_neq_components,islower,xinv_st,get_state,get_type,xinv_tp,find_xinv_group_coords,get_xunk_size_coname

from xinv.core.tools import find_overlap_coords,find_ilocs,find_unk_idx
import numpy as np
import xarray as xr
from xinv.core.logging import xinvlogger
from xinv.linalg.inplace import cholesky_inplace,dtrsm_inplace,dsyrk_inplace
from xinv.neq import zeros as neqzeros


def ireduce(dsneq,idx,keep=False):
    """
        Reduce Normal equation system using indexing
        Parameters:
        -----------
        dsneq: xarray.Dataset
            Dataset containing the normal equation system to be reduced
        idx: list or np.ndarray 
            index of the coordinates to reduce or keep in the system 

    """
    
    u_sz,unkdim=get_xunk_size_coname(dsneq)
    #compute the complementary index
    if keep:
        idxkeep=idx
        if idxkeep.dtype == bool:
            idxreduce=~idxkeep
        else:
            idxreduce=~np.isin(np.arange(u_sz),idxkeep)
    else:
        idxreduce=idx
        if idxreduce.dtype == bool:
            idxkeep=~idxreduce
        else:
            idxkeep=~np.isin(np.arange(u_sz),idxreduce)

    #pointers to the input normal equation system
    i_N,i_rhs,i_x0,i_ltpl,i_sigma0,i_nobs,i_npara=find_neq_components(dsneq)

    #
    
    if i_x0 is None:
        i_x0=xr.zeros_like(i_rhs)
    
    
    i_lower=islower(i_N)

    #copy the upper/lower triangular (needed to get correct cross-section matrices)
    if i_lower:
        i_N+=np.tril(i_N,-1).T
    else:
        i_N+=np.triu(i_N,1).T

    #create new coordinates for the output system
    outcoords={}
    for k,v in dsneq.coords.items():
        if k == unkdim:
            outcoords[k]=v[idxkeep]
        elif  get_type(v) in [xinv_tp.grp_id_co,xinv_tp.grp_seq_co]:

            #ignore these as they will already be included in the multiIndex unkdim coordinate
            continue
        else:
            outcoords[k]=v
    
    o_dsneq= neqzeros(rhsdims=i_rhs.dims,coords=outcoords,lower=i_lower)
    

    #pointers to the new normal equation system
    o_N,o_rhs,o_x0,o_ltpl,o_sigma0,o_nobs,o_npara=find_neq_components(o_dsneq)
    
    #copy the relevant data from the input system
    o_x0[()]=i_x0[{unkdim:idxkeep}]
    o_ltpl[()]=i_ltpl
    o_sigma0[()]=i_sigma0
    o_nobs[()]=i_nobs
    o_npara[()]=i_npara

    #initialize the rhs with copy of the input
    o_rhs[()]=i_rhs[{unkdim:idxkeep}]

    #initialize the Nkk with copy of the input
    o_N[()]=i_N[{unkdim:idxkeep,unkdim+'_':idxkeep}]

    #obtain the sections of the input normal system
    Nrr=i_N[{unkdim:idxreduce,unkdim+'_':idxreduce}]
    
    Nrk=i_N[{unkdim:idxreduce,unkdim+'_':idxkeep}]
    

    #retrieve the Cholesky decomposition of the to be reduced part
    
    cholesky_inplace(Nrr)
    
    #Decorrelate the Nrk matrix with Nrr cholesky part
    # that is solve Nrr' *X=Nrk  for X
    dtrsm_inplace(Nrr,Nrk,trans=1)
    
    
    #decorrelate the rhs part
    br=i_rhs[{unkdim:idxreduce}]
    dtrsm_inplace(Nrr,br,trans=1)
    
    o_ltpl[()]-=br.dot(br,dim=unkdim)
    #update right hand side
    o_rhs[()]-=Nrk.T.dot(br,dim=unkdim).rename({unkdim+'_':unkdim})
    
    #Update the normal matrix
    dsyrk_inplace(o_N,Nrk,trans=1,beta=1.0,alpha=-1.0)

    return o_dsneq

def reduce(dsneq, keep=False,**kwargs):
    """
        Reduce Normal equation system using coordinate labelling
        Parameters:
        -----------
        dsneq: xarray.Dataset
            Dataset containing the normal equation system to be reduced
        **kwargs: 
            keyword arguments with the dimension name as key and a list of coordinate labels to be reduced from the system
            coord1 = reducelabels1 , .. coord2 = reducelabels2 

    """
    
    idxfound,idxremaining,idxnotfound=find_unk_idx(dsneq,**kwargs)
    if idxnotfound is not None:
        xinvlogger.warning(f"Reduction parameters contain values {idxnotfound} which are not found in the input normal equation system, ignoring those")

    if (not keep and idxremaining is None) or (keep and idxfound is None):
        xinvlogger.warning("Nothing to reduce, returning input")
        return dsneq
    elif (not keep and idxfound is None) or (keep and idxremaining is None):
        #cannot reduce all unknowns
        raise ValueError("Reduction parameters contain all unknown parameters, cannot reduce all unknowns")
    
    return ireduce(dsneq,idxfound,keep)

def groupreduce(dsneq,groupname,keep=False):
    """
    Reduce by groupname a group of parameters from a normal equation system
    Parameters
    ----------
    dsneq : xr.Dataset
        Dataset containing the normal equation system to be reduced
    groupname : str
        The groupname of the parameters to be reduced
    keep : bool, optional
        If True, the group parameters are kept instead of reduced. The default is False.
    """


    #test whether the groupname actually exists
    grpid_co,grpseq_co=find_xinv_group_coords(dsneq)
    if grpid_co is None or grpseq_co is None:
        raise ValueError("No group coordinates found in the normal equation system")

    if groupname not in grpid_co.data:
        raise ValueError(f"Groupname {groupname} not found in the normal equation system")

    idx=grpid_co.data == groupname
    return ireduce(dsneq,idx,keep)





## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl
## Read Normal equation systems or solutions from SINEX (2.02) files
## This is currently not feature complete



import numpy as np
import xarray as xr
from datetime import datetime,timedelta
from xinv.core.grouping import build_group_coord
from xinv.core.logging import xinvlogger
from xinv.io.gzipwrap import gzip_open_r
from xinv.xiext import  read_symmat_fast
from xinv.core.attrs import * 
import pandas as pd

snxunkdim="snx_unk"

snxunkco=f"_{snxunkdim}_idx"

class SNXBlock:
    """
    Base class to parse various SINEX blocks
    """
    def __init__(self,blockname):
        self.block=blockname
    def parseblock(self,fileobj,dsout):
        pass



def sinex2date(snxdate:str)->datetime:
    """
        Convert sinex datestring in yy:doy:seconds to python datetime

    Parameters
    ----------
    snxdate : str
        datestring in sinex format
        

    Returns
    -------
    datetime
       Specified date and time as datetime object 

    """
    yr,doy,sec=[int(x) for x in snxdate.split(":")]
    return datetime(yr+2000 if yr<50 else yr+1900,1,1)+timedelta(days=doy-1,seconds=sec)


def read_symmat_slow(fileobj,dsout,blockname):
    """
        Reads a triangular matrix from a SINEX block and returns a symmetric version.
        This is a slow Pure Python implmentation, use read_symmat_fast for a faster cythonized version
    Parameters
    ----------
    fileobj : 
        io buffer to read lines from
        
    dsout : xarray.Dataset
        xarray.Dataset to augment the matrix data to
        
        
    blockname : str
        name of the SINEX block. Should be one of:
        SOLUTION/NORMAL_EQUATION_MATRIX U
        SOLUTION/NORMAL_EQUATION_MATRIX L
        

    Returns
    -------
    an updated xarray.Dataset holding the new matrix in a new variable 'N'

    """
    if not blockname.startswith('SOLUTION/NORMAL_EQUATION_MATRIX'):
        raise RuntimeError(f"Wrong block {blockname}?")
    nest,xunk_co=get_xunk_size_coname(dsout)
    xunkdim=dsout[xunk_co].dims[0]
    mat=np.zeros([nest,nest],order='C')
    data=np.zeros([3])
    for line in fileobj:
        if line[0] == '-':
            #end of block encountered
            break
        elif line[0] =='*':
            #comment
            continue
        data=[float(x) for x in line.split()]
        irow=int(data[0])-1 #note zero indexing
        icol=int(data[1])-1
        ndat=len(data)-2
        mat[irow,icol:icol+ndat]=data[2:]
    #mirror the upper triangle in the lower part
    mat=np.triu(mat,k=1).T+mat

    # if "nm_" not in dsout.indexes:
        # #add the transposed index
        # mi_=SHindexBase.mi_toggle(dsout.indexes['nm'])
        # dsout=dsout.sh.set_nmindex(mi_,'_')

    dsout['N']=([xunkdim,xunkdim+'_'],mat)
    dsout.N.attrs.update(N_attrs())
    return dsout

class SNXVEC(SNXBlock):
    def __init__(self,blockname):
        super().__init__(blockname)
        self._dispatch=dict(CN=self.CNparse,SN=self.CNparse,STAX=self.STA_parse,STAY=self.STA_parse,STAZ=self.STA_parse)
        self._coords={}
        self._coord_multindex_names={}
        if blockname == "SOLUTION/APRIORI":
            self.vname="apri_est"
            self.svname=None
            self.vattrs=x0_attrs()
            self.svattrs={}
        elif blockname == "SOLUTION/NORMAL_EQUATION_VECTOR":
            self.vname="rhs"
            self.svname=None
            self.vattrs=rhs_attrs()
            self.svattrs={}
        elif blockname == "SOLUTION/ESTIMATE":
            self.vname="sol_est"
            self.svname="sol_std"
            self.vattrs=solest_attrs()
            self.svattrs=stdsolest_attrs()
        else:
            raise NotImplementedError(f"{blockname} not recognized")


    def parseblock(self, fileobj, dsout):
         
        
        #allocate space
        nest,unkco=get_xunk_size_coname(dsout)
        unkdim=dsout[unkco].dims[0]
        #allocate space
        values=np.zeros([nest])
        if self.svname is not None:
            stddevs=np.zeros([nest])

        refepochs=np.empty([nest],dtype=datetime)
        grp_co=np.empty([nest],dtype=object)

        for line in fileobj:
            if line.startswith('-'):
                #end of block encountered
                break
            elif line.startswith('*'):
                #comment
                continue

            #find the parameter group, and line parser from the type
            snxtype=line[7:7+6].strip()
            idx,grpidseq,refepoch,val,std=self._dispatch.get(snxtype,self.generic_parse)(line)
            refepochs[idx]=refepoch
            values[idx]=val
            if std:
                stddevs[idx]=std
            grp_co[idx]=grpidseq
        dsout[self.vname]=(unkdim,values)
        dsout[self.vname].attrs.update(self.vattrs)
        
        if self.svname:
            dsout[self.svname]=(unkdim,stddevs)
            dsout[self.svname].attrs.update(self.svattrs)
        
        dsout['refepochs']=([unkdim],refepochs)
        
        #possibly add coordinates
        dsout=self.add_coords(dsout,grp_co)

        return dsout
    
    def add_coords(self,dsout,grp_co):
        
        nest,unkco=get_xunk_size_coname(dsout)
        unkdim=dsout[unkco].dims[0]
        if unkco != snxunkco:
            #no need to re-add coordinates, because tghey were already assigned in a prvious call of this function
            return dsout

        #change the state of the current unknown coordinate to unlinked
        change_state(dsout[snxunkco],xinv_st.unlinked)


        if len(self._coords) == 1:
            #only one group
            grpid=next(iter(self._coords))
            dsout=dsout.rename({unkdim:grpid})
            unkdim=grpid
            state=xinv_st.linked

        else:
            #multiple groups: add coordinates but in an unlinked state
            state=xinv_st.unlinked

        for grpid,coord in self._coords.items():

            if grpid  in ["nm","stat"]:
                #Modify the coordinateinto a multindex
                coord=pd.MultiIndex.from_tuples(coord, names=self._coord_multindex_names[grpid])
                coord=xr.Coordinates.from_pandas_multiindex(coord, grpid)
                dsout=dsout.assign_coords(coord)
                dsout[grpid].attrs.update(xunk_coords_attrs(state=state))
            else:
                #generic case
                dsout=dsout.assign_coords({grpid:(grpid,coord)})
                dsout[grpid].attrs.update(xunk_coords_attrs(state=state))
        
        #possibly add group_id and seq coordinates 
        if state == xinv_st.unlinked:
            dsout=dsout.assign_coords(build_group_coord(grp_co,dim=unkdim))

        return dsout

    def CNparse(self,line):
        grpid="nm"
        if grpid not in self._coords:
            #initialize subgroup
            self._coords[grpid]=[]
            self._coord_multindex_names[grpid]=["n","m"]

        fields = line.split()
        idx=int(fields[0])-1
        n = int(fields[2])
        m=int(fields[4])
        m=m if fields[1] == "CN" else -m
        
        self._coords[grpid].append((n,m))

        epoch=sinex2date(fields[5])
        val=float(fields[8])
        if self.svname:
            std=float(fields[9])
        else:
            std=None

         
        return idx,(grpid,len(self._coords[grpid])-1),epoch,val,std 
        

    def generic_parse(self,line):
        grpid="snxaux"

        if grpid not in self._coords:
            #initialize subgroup
            self._coords[grpid]=[]
            self._coord_multindex_names[grpid]=None
        idx=int(line[0:6])-1
        para=line[7:26]
        self._coords[grpid].append(para)
        epoch=sinex2date(line[27:27+12])
        val=float(line[46:46+22])

        if self.svname:
            std=float(line[69:69+22])
        else:
            std=None
        return idx,(grpid,len(self._coords[grpid])-1),epoch,val,std 


    def STA_parse(self,line):
        grpid="stat"

        if grpid not in self._coords:
            #initialize subgroup
            self._coords[grpid]=[]
            self._coord_multindex_names[grpid]=['statype','site','monument','soluid']
        idx=int(line[0:6])-1
        statype=line[7:7+6].strip()
        site=line[14:14+4]
        monument=line[19:19+2].strip()
        soluid=int(line[22:22+4])
                
        self._coords[grpid].append((statype,site,monument,soluid))
        epoch=sinex2date(line[27:27+12])
        val=float(line[46:46+22])

        if self.svname:
            std=float(line[69:69+22])
        else:
            std=None
        return idx,(grpid,len(self._coords[grpid])-1),epoch,val,std 




class STATISTICS(SNXBlock):
    def __init__(self,blockname):
        assert blockname == "SOLUTION/STATISTICS"
        super().__init__("SOLUTION/STATISTICS")

    def parseblock(self,fileobj,dsout):
        """
            Reads a SINEX block holding statistical metrics
        Parameters
        ----------
        fileobj : 
            io buffer to read lines from
            
        dsout : xarray.Dataset
            xarray.Dataset to augment the matrix data to
            
        Returns
        -------
        an updated xarray.Dataset holding the available statistics  as scalar variables


        """
        for line in fileobj:
            if line.startswith('-'):
                #end of block encountered
                break
            elif line.startswith('*'):
                #comment
                continue
                continue
            
            attrs=None

            if line.startswith(" NUMBER OF DEGREES OF FREEDOM"):
                varname="dof"
                tp=int
            elif line.startswith(" NUMBER OF OBSERVATIONS"):
                varname="nobs"
                tp=int
                attrs=nobs_attrs()
            elif line.startswith(" NUMBER OF UNKNOWNS"):
                varname="nunknown"
                tp=int
                attrs=npara_attrs()
            elif line.startswith(" WEIGHTED SQUARE SUM OF O-C"):
                varname="ltpl"
                tp=float
                attrs=ltpl_attrs()

            elif line.startswith(" VARIANCE FACTOR"):
                varname="sigma0_2"
                tp=float
                attrs=sigma0_attrs()
            elif line.startswith(" SQUARE SUM OF RESIDUALS (VTPV)"):
                varname="ltpl"
                tp=float
                attrs=ltpl_attrs()
            else:
                tp=None
                varname =None
                xinvlogger.warning(f"ignoring {self.block} entry {line}")

            if varname:
                spl = line.split()
                val=tp(spl[-1])
                if varname == "sigma0_2":
                    #convert to sigma0
                    val=np.sqrt(val)
                    varname="sigma0"
                dsout[varname]=val
                if attrs:
                    dsout[varname].attrs.update(attrs)
        return dsout


class FILEREF(SNXBlock):
    def __init__(self,blockname):
        assert blockname == "FILE/REFERENCE"
        super().__init__("FILE/REFERENCE")

    def parseblock(self,fileobj,dsout):
        globattrs={}
        for line in fileobj:
            if line.startswith('-'):
                #end of block encountered
                break
            elif line.startswith('*'):
                #comment
                continue
            ky=line[1:19]
            val=line[20:-1]
            globattrs["SNX-"+ky]=val
        dsout.attrs.update(globattrs)
        return dsout

class SYMMAT(SNXBlock):
    def __init__(self,blockname):
        super().__init__(blockname)

    def parseblock(self,fileobj,dsout):
        return read_symmat_fast(fileobj,dsout,self.block)
        # return read_symmat(fileobj,dsout,self.block)



# dictionary to lookup blockparser classes)

blockdispatch={"SOLUTION/STATISTICS":STATISTICS,
               "FILE/REFERENCE":FILEREF,
               "SOLUTION/APRIORI":SNXVEC,
               "SOLUTION/ESTIMATE":SNXVEC,
               "SOLUTION/NORMAL_EQUATION_VECTOR":SNXVEC,
               "SOLUTION/NORMAL_EQUATION_MATRIX U":SYMMAT,
               "SOLUTION/NORMAL_EQUATION_MATRIX L":SYMMAT}

compatversions=["2.02","2.01"]

def parse_headerline(line):
    if not line.startswith('%=SNX'):
        raise RuntimeError("Not a valid SINEX file")
    
    header={}
    header['version']=line[6:6+4]

    headerfields=line.split()
    if header['version'] not in compatversions:
        raise RuntimeError(f"read_sinex is not compatible with {header['version']}")
    
    header['nest']=int(line[60:60+5])

    header['tstart']=sinex2date(line[32:32+12])
    header['tend']=sinex2date(line[45:45+12])
    return header

def read_sinex(file_or_obj,stopatmat=False):
    """
        Reads normal equation information from a SINEX file (currently only partially supported)
    Parameters
    ----------
    file_or_obj : 
        IO buffer or filename with the SINEX data source
        
    stopatmat : bool
        Stop reading from the source when encountering a MATRIX block to speed up. (default=False)
        

    Returns
    -------
       a xarray.Dataset holding the normal equation information 
        

    """
    needsClosing=False
    if type(file_or_obj) == str:
        needsClosing=True
        if file_or_obj.endswith('.gz') or file_or_obj.endswith('.Z'): 
            file_or_obj=gzip_open_r(file_or_obj,textmode=True)
        else:
            file_or_obj=open(file_or_obj,'rt')
    
    line=file_or_obj.readline()
    header=parse_headerline(line)
    #initialize xarray dataset with some scalar vars to augment
    nest=header['nest']

    dsout=xr.Dataset(dict(tstart=header['tstart'],tend=header['tend']),coords={snxunkco:(snxunkdim,np.arange(nest))})
    # use  snx_ids as the unknown coordinate dimension (for now)
    dsout[snxunkco].attrs.update(xunk_coords_attrs(state=xinv_st.linked))
    
    # loop until a block is encountered and then dispatch to appropriate function
    for line in file_or_obj:
        if line[0] == "+":
            block=line[1:].strip()

            if stopatmat and "MATRIX" in block:
                xinvlogger.info(f"Encountered {block}, stopping")
                break
            if block not in blockdispatch.keys():
                xinvlogger.warning(f"Ignoring block {block}")
                continue
            xinvlogger.info(f"Reading block {block}")
            
            blockobj=blockdispatch[block](block)
            dsout=blockobj.parseblock(file_or_obj,dsout)

        
    if needsClosing:
        file_or_obj.close()

    if "sigma0" not in dsout:
        dsout["sigma0"]=xr.DataArray(1.0)
        dsout.sigma0.attrs.update(sigma0_attrs())

    return dsout 



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, rietbroek@utwente.nl

import xarray as xr
import numpy as np
from xinv.fwd import FwdOpbase
from xinv.core.grouping import expand_as_group

class FwdStackOp(FwdOpbase):
    def __init__(self,fwdops=None,cache=False,unknown_dim="xinv_unk"):
        """Setup a forward operator which consists of stacking several other forward operators as groups which share the observation dimension"""
        super().__init__(cache=cache,unknown_dim=unknown_dim)
        self._fwdops=[]
        if hasattr(fwdops,'__iter__'):
            for fwdop in fwdops:
                self.append(fwdop)
        else:
            self.append(fwdops)

    def _jacobian_impl(self,**kwargs):
        """Creates the Jacobian of the forward operator"""
        jacobian=None
        for stack_id,fwdop in enumerate(self._fwdops):
            #get Jacobian and add a multindex holding its stackid
            jacobian_i=fwdop.jacobian(**kwargs)
            jacobian_i=expand_as_group(jacobian_i,group_dim=fwdop._unkdim,stack_dim=self._unkdim)
            if jacobian is None:
                jacobian=jacobian_i
            else:
                #Not really memory friendly at the moment, but ok for now
                jacobian=xr.concat([jacobian,jacobian_i],dim=self._unkdim)
        return jacobian

    def append(self,fwdop):
        if not isinstance(fwdop,FwdOpbase):
            raise ValueError("Input should be a FwdOpbase instance or derived class")
        if len(self._fwdops) == 0:
            #take the observational dimension from the first forward operator
            self._obsdim=fwdop._obsdim

        self._fwdops.append(fwdop)
        



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import numpy as np
import xarray as xr
from scipy.linalg import cholesky

from xinv.cov import covariancebase
from xinv.cov.covariancebase import CovarianceBase

class CovarianceMat(CovarianceBase):
    def __init__(self,N_or_Cov="Normal or Covariance matrix",error_cov=True):

        if error_cov is True:
            cov=N_or_Cov
            N=xr.apply_ufunc(np.linalg.inv,cov)
        else:
            N=N_or_Cov
            cov=xr.apply_ufunc(np.linalg.inv,N)

        super().__init__(N=N)

        naux_,naux=N.dims

        
        self._Ncholesky=xr.apply_ufunc(cholesky,N,input_core_dims=[["nm","nm_"]],output_core_dims=[["nm","nm_"]],kwargs={"lower":True})


    
    def _decorrelate_impl(self,damat):        
        
        decorrelated=xr.dot(self._Ncholesky,damat)

        return decorrelated

class DiagonalCovarianceMat(CovarianceBase):
    def __init__(self,diag_std):

        self._Ncholesky=1/diag_std
        N=xr.apply_ufunc(np.square,self._Ncholesky) 
        super().__init__(N=N)
        
    def _decorrelate_impl(self,damat):
        

        decorrelated=self._Ncholesky*damat


        return decorrelated
        
    


## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl


from xinv.core.attrs import find_component,xinv_tp,find_xinv_group_coords,get_xunk_size_coname

from xinv.core.tools import find_unk_idx
import numpy as np
from xinv.core.logging import xinvlogger


def ifix(dsneq,idx,keep=False):
    """
        Fix/remove parameters in a normal equation system to their apriori values using indexing
        Parameters:
        -----------
        dsneq: xarray.Dataset
            Dataset containing the normal equation system to be fixed
        idx: list or np.ndarray 
            index of the coordinates to fix or keep in the system 

    """
    
    u_sz,unkdim=get_xunk_size_coname(dsneq)
    #compute the complementary index
    if keep:
        idxkeep=idx
        if idxkeep.dtype == bool:
            idxfix=~idxkeep
        else:
            idxfix=~np.isin(np.arange(u_sz),idxkeep)
    else:
        idxfix=idx
        if idxfix.dtype == bool:
            idxkeep=~idxfix
        else:
            idxkeep=~np.isin(np.arange(u_sz),idxfix)

    o_dsneq=dsneq.sel({unkdim:idxkeep,unkdim+'_':idxkeep})
    io_npara=find_component(o_dsneq,xinv_tp.npara)
    #update amount of unknown parameters
    io_npara[()]-=len(idxfix)

    return o_dsneq


def fix(dsneq, keep=False,**kwargs):
    """
        Fix/remove parameters from a Normal equation system using coordinate labelling
        Parameters:
        -----------
        dsneq: xarray.Dataset
            Dataset containing the normal equation system to be fixed
        keep: bool, optional
            If True, the parameters are kept instead of fixed. The default is False.
        **kwargs: 
            keyword arguments with the dimension name as key and a list of coordinate labels to be fixed/removed from the system
            coord1 = fixlabels1 , .. coord2 = fixlabels2 

    """
    
    idxfound,idxremaining,idxnotfound=find_unk_idx(dsneq,**kwargs)
    if idxnotfound is not None:
        xinvlogger.warning(f"Fix parameters contain values {idxnotfound} which are not found in the input normal equation system, ignoring those")

    if (not keep and idxremaining is None) or (keep and idxfound is None):
        xinvlogger.warning("Nothing to fix, returning input")
        return dsneq
    elif (not keep and idxfound is None) or (keep and idxremaining is None):
        #cannot fix all unknowns
        raise ValueError("Fix parameters contain all unknown parameters, cannot fix all unknowns")
    
    return ifix(dsneq,idx=idxfound,keep=keep)

def groupfix(dsneq,groupname,keep=False):
    """
    Fix/remove by groupname a group of parameters from a normal equation system
    Parameters
    ----------
    dsneq : xr.Dataset
        Dataset containing the normal equation system to be fixed/removed
    groupname : str
        The groupname of the parameters to be fixed/removed
    keep : bool, optional
        If True, the group parameters are kept instead of fixed. The default is False.
    """


    #test whether the groupname actually exists
    grpid_co,grpseq_co=find_xinv_group_coords(dsneq)
    if grpid_co is None or grpseq_co is None:
        raise ValueError("No group coordinates found in the normal equation system")

    if groupname not in grpid_co.data:
        raise ValueError(f"Groupname {groupname} not found in the normal equation system")

    idx=grpid_co.data == groupname
    return ifix(dsneq,idx=idx,keep=keep)





# This file is part of the shxarray software which is licensed
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

#
# distutils: language = c++
# cython: profile=False
"""
**xiext** is xinv  binary Cython backend. 
Some of the heavy lifting is done using this the functions of this shared library.
This file groups cython functionality under a common shared library
"""

include "sinex.pyx" 



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

class XinvIllposedError(Exception):
    """Exception raised when Cholesky decomposition is ill-posed 

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, linalgmessage):
        self.message = f"Cholesky decomposition failed: {linalgmessage}"
        super().__init__(self.message)




#!/usr/bin/env python3
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv import *
from xinv.fwd.polynomial import Polynomial
from xinv.fwd.harmonics import SeasonalHarmonics
from xinv.fwd.fwdstack import FwdStackOp
from xinv.core.exceptions import XinvIllposedError
import os

neqfile1=os.path.join(os.path.dirname(__file__),f'testdata/neqpoly.nc')

#note apply a seed to garantee reproducibility (otherwise tests may fail in  statistical sense)
rg=np.random.default_rng(12789)

@pytest.fixture(params=['C','F'])
def noisystacked(request):
    """
    Generate a noisy dataset from a set of polynomial functions and annual and semi-annual signals
    """

    npoly=2
    noise_std=0.01
    naux=5 #number of auxiliary datasets
    dtype='datetime64[D]'
    t_axis=np.arange('2001-01-31', '2020-12-31', dtype=dtype)
    t0=np.datetime64("2010-01-01",'D') 
    delta_t=np.timedelta64(365,'D')+ np.timedelta64(6,'h')
    order=request.param[0]
    obs=np.zeros([naux,len(t_axis)],order=order)
    
    t_axis_rel=(t_axis-t0)/delta_t #relative time axis
    #add polynomial truth values
    polytrue=np.zeros([naux,npoly+1])
    for i in range(naux):
        polytrue[i,:]=[j*i for j in range(npoly+1)]
        obs[i,:]=np.polyval(polytrue[i,::-1],t_axis_rel)
    
    nharm=2
    seastrue=np.zeros([naux,nharm*2])
    omegas=[2*np.pi,4*np.pi]
    frac=0.25 #fraction of the amplitude
    #add seasonal harmonics
    for i in range(naux):
        #set cosine and sine amplitudes
        camp=naux
        samp=1-naux
        scamp=frac*samp
        ssamp=frac*samp
        seastrue[i,:]= [camp,samp,ssamp,scamp] 
        obs[i,:]+=camp*np.cos(omegas[0]*t_axis_rel)+samp*np.sin(omegas[0]*t_axis_rel)+scamp*np.cos(omegas[1]*t_axis_rel)+ssamp*np.sin(omegas[1]*t_axis_rel)

    #add some normal noise
    obs+= rg.normal(0,noise_std,obs.shape)
    
    # create a naming of the auxdims
    auxcoord=[f"aux_{i}" for i in range(naux)]
    dsobs=xr.Dataset({"obs":xr.DataArray(obs,dims=("naux","time")),"polytrue":xr.DataArray(polytrue.T,dims=("poly","naux")),"seastrue":xr.DataArray(seastrue.T,dims=("seas","naux"))},coords={"time":t_axis,"naux":auxcoord},attrs=dict(noise_std=noise_std,t0=t0,delta_t=delta_t,npoly=npoly))

    return dsobs


def test_illposed(noisystacked):
    """
    Test a stacked forward operator, consisting of multiple stacked forward operators, building of a normal equation system, and solving step. This setup shoudl results in an illposed system and should be captured as such
    Parameters
    ----------
    noisystacked : xr.Dataset containing the noisy polynomial observations and the true polynomial coefficients

    """
    #initialize the first polynomial forward operator
    npoly=noisystacked.attrs['npoly']
    t0=noisystacked.attrs['t0']
    delta_t=noisystacked.attrs['delta_t']
    polyfwd=Polynomial(n=npoly,poly_x='time',x0=t0,delta_x=delta_t,cache=True)
    
    #initialize the second polynomial forward operator (n=0, so illposed because its also contained within polyfwd)
    polyfwd2=Polynomial(n=0,poly_x='time',x0=t0,delta_x=delta_t,unknown_dim="poly2")
    #initialize the stacked forward operator
    fwdstck=FwdStackOp(polyfwd)
    fwdstck.append(polyfwd2)

    # #build the normal equation system
    std_noise=0.5
    dsneq=noisystacked.obs.xi.build_normal(fwdstck,ecov=std_noise*std_noise) 

    try:
        dssol=dsneq.xi.solve()
        #we're not supposed to end up here
        assert False
    except XinvIllposedError as e:
        assert True
        
def test_stacked(noisystacked):
    """
    Test a stacked forward operator, consisting of multiple stacked forward operators, building of a normal equation system, and solving step. 
    Parameters
    ----------
    noisystacked : xr.Dataset containing the noisy polynomial observations and the true polynomial coefficients and harmonic coefficients to test against

    """

    #initialize the first polynomial forward operator
    npoly=noisystacked.attrs['npoly']
    t0=noisystacked.attrs['t0']
    delta_t=noisystacked.attrs['delta_t']
    polyfwd=Polynomial(n=npoly,poly_x='time',cache=True,x0=t0,delta_x=delta_t)
    
    #initialize the stacked forward operator
    fwdstck=FwdStackOp(polyfwd)

    #Append Annual and Seminannual fwd operators
    seasfwd=SeasonalHarmonics(x0=t0,semi_annual=True)
    fwdstck.append(seasfwd)

    # #build the normal equation system
    std_noise=0.5
    dsneq=noisystacked.obs.xi.build_normal(fwdstck,ecov=std_noise*std_noise)
    if not os.path.exists(neqfile1):
        #write the normal equation system to a file (used for other tests)
        dsneq.reset_index('xinv_unk').to_netcdf(neqfile1)

    dssol=dsneq.xi.solve()

    #extract the groups of the solution and compare to the true values
    dsolpoly=dssol.xi.get_group('poly')
    prenoise=noisystacked.attrs['noise_std']
    tol=3*np.sqrt(np.diag(dsolpoly.COV))*prenoise
    tol=tol.max()
    assert np.allclose(dsolpoly.solution,noisystacked.polytrue,atol=tol)

    dsolseas=dssol.xi.get_group('harmonics_seasonal')
    tol=3*np.sqrt(np.diag(dsolseas.COV))*prenoise
    tol=tol.max()
    assert np.allclose(dsolseas.solution,noisystacked.seastrue,atol=tol)



# This file is part of the shxarray software which is licensed
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl


from gzip import GzipFile
import os
from io import TextIOWrapper,BytesIO

try:
    from rapidgzip import RapidgzipFile
except:
    #failed to load (ok)
    RapidgzipFile=None

def gzip_open_r(filename,textmode=False):
    """
        GZip file reading wrapper leveraging parallel decompression speed when rapidgzip is installed on the system
    Parameters
    ----------
    filename : str
        Filename of the gzip archive to open
        
    textmode : 
        Whether to open in textmode (allows iterating over lines)
        
    """
    
    if filename.endswith('.Z'):
        
        try:
            import unlzw3
        except ImportError:
            raise ImportError("unlzw3 is not installed, please install it to use the .Z file format")
        #use uncompress
        with open(filename, 'rb') as f:
            #rtead entire file
            gzfid = BytesIO(unlzw3.unlzw(f.read()))


    elif RapidgzipFile: 
        gzfid=RapidgzipFile(filename,parallelization=os.cpu_count())

    else:
        gzfid=GzipFile(filename,'rb')


    if textmode:
        return TextIOWrapper(gzfid)
    else:
        return gzfid






## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

from scipy.linalg import cholesky
import numpy as np
import xarray as xr

from scipy.linalg.blas import dtrsm
from scipy.linalg.lapack import dpotri
from xinv.core.exceptions import XinvIllposedError

from xinv.core.attrs import cov_attrs, solest_attrs,ltpl_attrs,find_neq_components,sigma0_attrs,Chol_attrs,xinv_st,islower

from xinv.linalg.inplace import cholesky_inplace, dpotri_inplace,dtrsm_inplace

def solve(dsneq,inplace=False):
    """Solve the normal equation system using Cholesky factorization"""
    
    if not inplace:
        #copy entire NEQ and operate on that one in place
        dsneq=dsneq.copy(deep=True)


    N,rhs,x0,ltpl,sigma0,nobs,npara=find_neq_components(dsneq)

    b=rhs.copy(deep=True)
    
    #decompose the normal matrix using cholesky in place N -> U'U or LL'
    cholesky_inplace(N)
    
    
    #Solve the system in several steps
    #1. triangular solve U'z=rhs for z (U' is now the Cholesky factor of N)
   
    dtrsm_inplace(N,rhs,trans=1)

    #2 solve Ux=z for x
    
    dtrsm_inplace(N,rhs)
    
    #3. update ltpl
    ltpl-=rhs.dot(b,dim=N.dims[0])
    ltpl.attrs.update(ltpl_attrs('posteriori'))
    
    # dtrsv(N.data,rhs.data,lower=0,trans=0,overwrite_x=1)
    rhs.attrs.update(solest_attrs())


    #4 compute error-covariance 
    # call dpotri('U',n1,C,n1,info)
    dpotri_inplace(N)
    # compute posteriori sigma0
    
    sigma0=np.sqrt(ltpl/(nobs-npara))
    sigma0.attrs.update(sigma0_attrs('posteriori'))
    dsneq['sigma0']=sigma0
    
    dsneq=dsneq.rename(dict(N='COV',rhs='solution'))
    if inplace:
        return None
    else:
        return dsneq



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv import *
import os

from fixtures import neqbase
from xinv.core.tools import find_ilocs


@pytest.mark.parametrize("neqbase",["simple","stacked"],indirect=True)
@pytest.mark.parametrize("keep",[True,False])
def test_reduce(neqbase,keep):
    """
    Test the reduction of a set of unknown parameters from a neq system 
    """
    reducepolyparam=[2]
    if "harmonics_seasonal" in neqbase.coords:
        neqbase_red=neqbase.xi.reduce(poly=reducepolyparam,harmonics_seasonal=[2],keep=keep)
        npara=len(reducepolyparam)+1
    else:
        neqbase_red=neqbase.xi.reduce(poly=reducepolyparam,keep=keep)
        npara=len(reducepolyparam)
    

    dssolred=neqbase_red.xi.solve()

    #check if the correct variables were removed kept

    dssol=neqbase.xi.solve()
    unkdim=neqbase.xi.unknown_dim()
    #select the same parameters from the non-reduced solution
    idxkeep=find_ilocs(dssol,unkdim,dssolred[unkdim])
    if keep:
        assert len(idxkeep) == npara
    else:
        assert neqbase.sizes[unkdim]-len(idxkeep) == npara

    dssolsub=dssol.isel({unkdim:idxkeep,unkdim+"_":idxkeep})

    #solution should be the same on the non-reduced neq system
    assert np.allclose(dssolred.solution,dssolsub.solution)

    assert np.allclose(dssolred.COV,dssolsub.COV)
    assert np.allclose(dssolred.ltpl,dssolsub.ltpl)
    assert np.allclose(dssolred.sigma0,dssolsub.sigma0)

    

@pytest.mark.parametrize("neqbase",["simple","stacked"],indirect=True)
@pytest.mark.parametrize("keep",[True,False])
def test_groupreduce(neqbase,keep):
    """
    Test the reduction of a  group of unknown parameters from a neq system 
    """
    try:
        neqbase_red=neqbase.xi.groupreduce('poly',keep=keep)

    except ValueError:
        #capture errors when there are not dedicated groups
        if "harmonics_seasonal" not in neqbase.coords:
            #ok just return
            assert True
            return


    npara=neqbase.sizes['poly']
    


    dssolred=neqbase_red.xi.solve()

    #check if the correct variables were removed kept

    dssol=neqbase.xi.solve()
    unkdim=neqbase.xi.unknown_dim()
    #select the same parameters from the non-reduced solution
    idxkeep=find_ilocs(dssol,unkdim,dssolred[unkdim])
    if keep:
        assert len(idxkeep) == npara
    else:
        assert neqbase.sizes[unkdim]-len(idxkeep) == npara

    dssolsub=dssol.isel({unkdim:idxkeep,unkdim+"_":idxkeep})

    #solution should be the same on the non-reduced neq system
    assert np.allclose(dssolred.solution,dssolsub.solution)

    assert np.allclose(dssolred.COV,dssolsub.COV)
    assert np.allclose(dssolred.ltpl,dssolsub.ltpl)
    assert np.allclose(dssolred.sigma0,dssolsub.sigma0)




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv.cov.cov import CovarianceMat,DiagonalCovarianceMat
from xinv.fwd.polynomial import Polynomial
import os
#from xinv import *

rg=np.random.default_rng()

@pytest.fixture(params=['C','F'])
def noisypoly(request):
    """ 
    Generate a noisy dataset from a set of polynomial functions
    """
    npoly=3
    noise_std=0.01
    x0=0
    delta_x=2
    naux=10
    x_axis=np.arange(-2,3,0.05)
    x_axis_rel=(x_axis-x0)/delta_x
    order=request.param[0]
    polyobs=np.zeros([naux,len(x_axis)],order=order)
    polytrue=np.zeros([naux,npoly+1])
    damat=np.stack([x_axis_rel**i for i in range(npoly+1)],axis=1)

    for i in range(naux):
        polytrue[i,:]=[j*i for j in range(npoly+1)]
        polyobs[i,:]=np.polyval(polytrue[i,::-1],x_axis_rel)

    polyobs+=rg.normal(0,noise_std,polyobs.shape)

    # create a naming of the auxdims
    auxcoord=[f"aux_{i}" for i in range(naux)]
    dspoly=xr.Dataset({"polyobs":xr.DataArray(polyobs,dims=("naux","nm")),"polytrue":xr.DataArray(polytrue,dims=("naux","poly")),"damat":xr.DataArray(damat,dims=("nm","poly"))},coords={"naux":auxcoord,"nm":x_axis,"poly":np.arange(npoly+1)},attrs=dict(noise_std=noise_std,x0=x0,delta_x=delta_x))


    return dspoly


def poly_covariance(noisypoly):
    """
    Build a diagonal covariance matrix and standard deviation vector
    """
    std=noisypoly.damat.std(dim="poly")
    var=std**2
    
    std_da=xr.DataArray(std.data,dims="nm",coords={"nm":noisypoly.nm})
    cov_da=xr.DataArray(np.diag(var.data),dims=("nm","nm_"),coords={"nm":noisypoly.nm.data,"nm_":noisypoly.nm.data})

    return cov_da,std_da




def test_poly_decorrelation(noisypoly):
    """
    Compare decorrelation using a full diagonal covariance matrix vs. diagonal std-based decorrelation
    """
    cov_da,std_da=poly_covariance(noisypoly)

    cov_full=CovarianceMat(N_or_Cov=cov_da,error_cov=True)
    decorr_full=cov_full.decorrelate(noisypoly.damat).decorrelated
    
    cov_diag=DiagonalCovarianceMat(diag_std=std_da)
    decorr_diag=cov_diag.decorrelate(noisypoly.damat).decorrelated


    assert np.allclose(decorr_full.data,decorr_diag,atol=1e-15)




    



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from datetime import datetime
import os

project = 'xinv'
copyright = str(datetime.now().year)+', Roelof Rietbroek'
author = 'Roelof Rietbroek'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = ['nbsphinx', 'sphinxcontrib.apidoc', 'sphinx.ext.autodoc','sphinx.ext.napoleon']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store','**.ipynb_checkpoints']

apidoc_template_dir='_templates'
#figure out the actual installation directory
import xinv
apidoc_module_dir=os.path.dirname(xinv.__file__)

apidoc_output_dir = 'references'
apidoc_separate_modules = True
apidoc_module_first=True
apidoc_toc_file=False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
#html_theme = 'sphinx_rtd_theme'
html_theme='pydata_sphinx_theme'

html_static_path = ['_static']

napoleon_numpy_docstring = True

nbsphinx_prolog = """
Download this Jupyter notebook from `github <https://github.com/strawpants/xinv/blob/main/docs/source/notebooks/{{ env.doc2path(env.docname, base=None) }}>`_

----
"""
html_theme_options = {
    "use_edit_page_button": False,
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/strawpants/xinv",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
   ],
   "logo": {
        "alt_text": "xinv - Home",
        "image_light": "_static/xinv_logo_wtext_light.png",
        "image_dark": "_static/xinv_logo_wtext_dark.png",
    }
}

html_favicon= '_static/favicon.ico'

# html_context = {
    # # "github_url": "https://github.com", # or your GitHub Enterprise site
    # "github_user": "ITC-Water-Resources",
    # "github_repo": "shxarray",
    # "github_version": "main",
    # "doc_path": "docs/source",
# }


# def linkcode_resolve(domain, info):
    # if domain != 'py':
        # return None
    # if not info['module']:
        # return None
    # filename = info['module'].replace('.', '/')
    # return "https://github.com/ITC-Water-Resources/shxarray/%s.py" % filename

# linkcode_url="https://github.com/ITC-Water-Resources/shxarray"
# linkcode_link_text="view on"



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import xarray as xr

class CovarianceBase:
    def __init__(self,N=None,cache=False,decorrname="decorrelated"):
      #  self._cov=None
        self._N=N
        self._cache_decorrelate=cache
        self._decorr=None
        self._decorrname=decorrname

    def decorrelate(self,damat=None):
        """Decorrelate the damat using cholesky of the normal matrix"""
        
        if self._N is None and damat is None:
            raise ValueError("Requesting the decorrelated damat requires a valid covariance matrix")
        elif damat is None:
            return self._decorr



        decorr=self._decorrelate_impl(damat)
        if type(decorr) == xr.DataArray:
            decorr.name=self._decorrname
            decorr=decorr.to_dataset()

        if self._cache_decorrelate:
            self._decorr=decorr
            
        return decorr



#!/usr/bin/env python3
## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv import *
import os
from xinv.core.tools import find_ilocs
from xinv.core.attrs import find_xinv_coords,xinv_tp,xinv_st

from fixtures import neqbase

@pytest.mark.parametrize("neqbase",["simple","stacked"],indirect=True)
@pytest.mark.parametrize("auxdim",["allaux","aux_0"])
def test_addneqs_same(neqbase,auxdim):
    """
    Test the adding of 2 normal equation systems with the same overlap of parameters 
    """
    if auxdim != "allaux":
        neqbase=neqbase.sel(naux=auxdim)

    neqbase2=neqbase.copy(deep=True)
    #add the two normal equation systems (both are the same)
    neqcomb_same=neqbase.xi.add(neqbase2)
   
    unk_xinvcoords=find_xinv_coords(neqbase,include=[xinv_tp.unk_co],state=xinv_st.linked)
    #Should be one only
    unk_name=next(iter(unk_xinvcoords.keys()))


    #note combined normal equation system may have a different sorting
    idx1=find_ilocs(neqcomb_same,unk_name,neqbase.coords[unk_name].data)
    
    #parameters should stay the same as there are no implicitly reduced parameters
    assert np.allclose(neqcomb_same.npara,neqbase.npara)
    #number of observations should be doubled
    assert np.allclose(neqcomb_same.nobs,2*neqbase.nobs)
    
    
    #Normal matrix should be twice the base input
    assert np.allclose(neqcomb_same.N[idx1,idx1],2*neqbase.N)

    #Right hand side should be twice the base input
    assert np.allclose(neqcomb_same.rhs[{unk_name:idx1}],2*neqbase.rhs)

    #LtpL should be twice the base input
    assert np.allclose(neqcomb_same.ltpl,2*neqbase.ltpl)
    
    #Sigma should be the same as the first base input
    assert np.allclose(neqcomb_same.sigma0,neqbase.sigma0)


@pytest.mark.parametrize("neqbase",["stacked"],indirect=True)
def test_addneq_compl(neqbase):
    """ 
    test adding two normal equation systems with a partial overlap of parameters
    """    
    neqbase2=neqbase.copy(deep=True)
    #modify the second normal equation system so that it has a different set of parameters but with an overlap
    neqbase2=neqbase2.xi.rename_groups(dict(poly='poly2'))


    #add the two normal equation systems (both are the same)
    neqcomb=neqbase.xi.add(neqbase2)
   
    #LtpL should be the sum of the base inputs
    assert np.allclose(neqcomb.ltpl,neqbase.ltpl+neqbase2.ltpl)
    
    #Sigma should be the same as the first base input
    assert np.allclose(neqcomb.sigma0,neqbase.sigma0)

    #number of parameters should be 10
    assert np.allclose([10]*neqcomb.sizes['naux'],neqcomb.npara)

    assert np.allclose(neqcomb.nobs,neqbase.nobs+neqbase2.nobs)
    
    #system 1 subset test
    neqc_poly=neqcomb.xi.get_group('poly')
    ixpoly=find_ilocs(neqbase,'poly',neqc_poly.poly.data)
    #polynomial coefficient entries should mathc the original ones of the first system
    assert np.allclose(neqbase.N[ixpoly,ixpoly],neqc_poly.N)
    assert np.allclose(neqbase.rhs[{'xinv_unk':ixpoly}],neqc_poly.rhs)

    #system 2 subset test
    neqc_poly=neqcomb.xi.get_group('poly2')
    ixpoly=find_ilocs(neqbase2,'poly2',neqc_poly.poly.data)
    #polynomial coefficient entries should mathc the original ones of the first system
    assert np.allclose(neqbase2.N[ixpoly,ixpoly],neqc_poly.N)
    assert np.allclose(neqbase2.rhs[{'xinv_unk':ixpoly}],neqc_poly.rhs)

    
    #check overlapping parameters
    neqc_seas=neqcomb.xi.get_group('harmonics_seasonal')
    neq1_seas=neqbase.xi.get_group('harmonics_seasonal')
    neq2_seas=neqbase2.xi.get_group('harmonics_seasonal')

    ixseas1=find_ilocs(neq1_seas,'harmonics_seasonal',neqc_seas.harmonics_seasonal.data)
    ixseas2=find_ilocs(neq2_seas,'harmonics_seasonal',neqc_seas.harmonics_seasonal.data)
    #Overlap should be the sum of the 2 systems
    assert np.allclose(neqc_seas.N,neq1_seas.N[ixseas1,ixseas1]+neq2_seas.N[ixseas2,ixseas2])
    
    assert np.allclose(neqc_seas.rhs,neq1_seas.rhs[{'harmonics_seasonal':ixseas1}]+neq2_seas.rhs[{'harmonics_seasonal':ixseas2}])




@pytest.mark.parametrize("neqbase",["simple"],indirect=True)
def test_addneqs_sigma(neqbase):
    """
    Test the adding of 2 normal equation systems with different apriori sigma0 
    """
    neqbase2=neqbase.copy(deep=True)


    # test for impossible sigma scaling (should fail)
    neqbase2['sigma0'][2]=0.1 #change only one value
    try:
        neqcomb=neqbase.xi.add(neqbase2)
        assert False, "Adding function should have raised an exception"
    except ValueError as exp:
        #ok, proceed with testing
        pass

    #modify sigma's of the second syst
    sigma_factor=2
    neqbase2['sigma0'][:]=sigma_factor*neqbase.sigma0
    
    pscale=1+1/(sigma_factor**2)

    #add the two normal equation systems (both are the same)
    neqcomb=neqbase.xi.add(neqbase2)
   
    unk_xinvcoords=find_xinv_coords(neqbase,include=[xinv_tp.unk_co],state=xinv_st.linked)
    # #Should be one only
    unk_name=next(iter(unk_xinvcoords.keys()))
    

    # #note combined normal equation system may have a different sorting
    idx1=find_ilocs(neqcomb,unk_name,neqbase.coords[unk_name].data)
    
    # #parameters should stay the same as there are no implicitly reduced parameters
    assert np.allclose(neqcomb.npara,neqbase.npara)
    # #number of observations should be doubled
    assert np.allclose(neqcomb.nobs,2*neqbase.nobs)
    
    
    # #Normal matrix should be pscale times the base input
    assert np.allclose(neqcomb.N[idx1,idx1],pscale*neqbase.N)

    # #Right hand side should be pscale times the base input
    assert np.allclose(neqcomb.rhs[{unk_name:idx1}],pscale*neqbase.rhs)

    # #LtpL should be pscale the base input
    assert np.allclose(neqcomb.ltpl,pscale*neqbase.ltpl)
    
    # #Sigma should be the same as the first base input
    assert np.allclose(neqcomb.sigma0,neqbase.sigma0)




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import pytest
import xarray as xr
import numpy as np
from xinv import *
import os

from fixtures import neqbase
from xinv.core.tools import find_ilocs,find_unk_idx
from xinv.core.logging import xinvlogger

@pytest.mark.parametrize("neqbase",["simple","stacked"],indirect=True)
@pytest.mark.parametrize("keep",[True])
def test_aprifix(neqbase,keep):
    """
    Test the apriori setting and removal of a set of unknown parameters from a neq system 
    """
    
    
    #compute the overall solution
    dssol=neqbase.xi.solve()

    #get the index vector corresponding to some polynomials to fix (or keep)

    fixpolyparam=[1,2]
    idxfix,idxremain,_=find_unk_idx(dssol,poly=fixpolyparam)
    unkdim=dssol.xi.unknown_dim()
    #set the apriori values of the unknown parameters to that from the overall solution
    if keep:
        dsneq_apriset=neqbase.xi.set_x0(dssol.solution[{unkdim:idxremain}],inplace=False)
    else:
        dsneq_apriset=neqbase.xi.set_x0(dssol.solution[{unkdim:idxfix}],inplace=False)
    
    
    dssolapr=dsneq_apriset.xi.solve()

    #check if x0+solution is the same as the overall solution


    #solution should be the same on the non-reduced neq system
    assert np.allclose(dssolapr.solution+dssolapr.x0,dssol.solution)

    assert np.allclose(dssolapr.COV,dssol.COV)
    assert np.allclose(dssolapr.ltpl,dssol.ltpl)
    assert np.allclose(dssolapr.sigma0,dssol.sigma0)

    # #fix the relevant parameters to their apriori values
    dsneq_aprifix=dsneq_apriset.xi.fix(poly=fixpolyparam,keep=keep)
    
    dssolfix=dsneq_aprifix.xi.solve()
    if keep:
        dssolsub=dssol.isel({unkdim:idxfix,unkdim+"_":idxfix})
    else:
        dssolsub=dssol.isel({unkdim:idxremain,unkdim+"_":idxremain})
    soldif=np.abs(dssolfix.solution+dssolfix.x0-dssolsub.solution)
    # stddev=xr.DataArray(np.sqrt(np.diag(dssolfix.COV)),dims=unkdim,coords={unkdim:dssolfix[unkdim]})*dssolfix.sigma0
    
    assert np.all(soldif < 1e-10)
    



@pytest.mark.parametrize("neqbase",["simple"],indirect=True)
def test_aprifix_singlerhs(neqbase):
    """
    Test the apriori setting and removal of a set of unknown parameters from a neq system (single dimension rhs version) 
    """
    
    #only select one single rhs
    neqbase=neqbase.sel(naux='aux_1')


    
    #compute the overall solution
    dssol=neqbase.xi.solve()

    #get the index vector corresponding to some polynomials to fix (or keep)

    fixpolyparam=[1,2]
    idxfix,idxremain,_=find_unk_idx(dssol,poly=fixpolyparam)

    unkdim=dssol.xi.unknown_dim()

    dsneq_apriset=neqbase.xi.set_x0(dssol.solution[{unkdim:idxfix}])
    

    dssolapr=dsneq_apriset.xi.solve()

    #check if x0+solution is the same as the overall solution


    #solution should be the same on the non-reduced neq system
    assert np.allclose(dssolapr.solution+dssolapr.x0,dssol.solution)

    assert np.allclose(dssolapr.COV,dssol.COV)
    assert np.allclose(dssolapr.ltpl,dssol.ltpl)
    assert np.allclose(dssolapr.sigma0,dssol.sigma0)

    # #fix the relevant parameters to their apriori values
    dsneq_aprifix=dsneq_apriset.xi.fix(poly=fixpolyparam)
    
    dssolfix=dsneq_aprifix.xi.solve()
    dssolsub=dssol.isel({unkdim:idxremain,unkdim+"_":idxremain})
    soldif=np.abs(dssolfix.solution+dssolfix.x0-dssolsub.solution)
    # stddev=xr.DataArray(np.sqrt(np.diag(dssolfix.COV)),dims=unkdim,coords={unkdim:dssolfix[unkdim]})*dssolfix.sigma0
    
    assert np.all(soldif < 1e-10)

    

@pytest.mark.parametrize("neqbase",["stacked"],indirect=True)
@pytest.mark.parametrize("keep",[False,True])
def test_groupfix(neqbase,keep):
    """
    Test the apriori fixing and removal of a  group of unknown parameters from a neq system 
    """
    

    #compute the overall solution
    idxpoly=neqbase.xinv_grp_id == "poly" 
    
    dssol=neqbase.xi.solve()
    
    unkdim=dssol.xi.unknown_dim()
    if keep:
        dsneq_apriset=neqbase.xi.set_x0(dssol.solution)
    else:
        dsneq_apriset=neqbase.xi.set_x0(dssol.solution[{unkdim:idxpoly}])

    neqbase_aprifix=dsneq_apriset.xi.groupfix('poly',keep=keep)


    npara=neqbase.sizes['poly']
    


    dssolapri=neqbase_aprifix.xi.solve()


    if keep:
        assert dssolapri.sizes[unkdim] == npara
    else:
        assert neqbase.sizes[unkdim]-dssolapri.sizes[unkdim] == npara
    if keep:
        soldif=np.abs(dssolapri.solution+dssolapri.x0-dssol.solution[idxpoly])
    else:
        soldif=np.abs(dssolapri.solution+dssolapri.x0-dssol.solution[~idxpoly])
    
    assert np.all(soldif < 1e-10)



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import numpy as np
from scipy.linalg.blas import dsyrk
from xinv.core.attrs import rhs_attrs,N_attrs,ltpl_attrs,sigma0_attrs,nobs_attrs,npara_attrs,find_xinv_coords, xunk_coords_attrs,aux_coords_attrs,x0_attrs

import xarray as xr

def build_normal(fwdop,daobs,ecov=1,**kwargs):
    
    if len(daobs.sizes) > 2:
        raise NotImplementedError("Cannot handle multiple auxiliary dimensions at the moment")
    

    
    #create a design matrix for the given observations (note pass the observation data to the jacobian implementation, where it may or may not be used)
    dsdesign=fwdop.jacobian(daobs=daobs,**kwargs)
     
    dadesign=dsdesign[fwdop.jacobname]
    obsdim=fwdop.obs_dim
    unkdim=fwdop.unkdim
    if daobs.dims[0] != obsdim:
        #garantee that the observation dimension is the first dimension
        daobs=daobs.T

    nrhs=daobs.shape[1]
    nrhsdim=daobs.dims[1]

    if not np.isscalar(ecov):
        #decorrelate  error covariance
        dadesign=ecov.decorrelate(dadesign) 
        daobs=ecov.decorrelate(daobs)
        sigma0=np.ones(nrhs)
    else:        
        sigma0=np.ones(nrhs)*ecov
    
    #create the normal equation system (not only upper triangle filled
    lower=0
    trans=1
    if dadesign.data.flags['C_CONTIGUOUS'] and dadesign.data.strides[1] == 8:
        pass
    elif dadesign.data.flags['F_CONTIGUOUS'] and dadesign.data.strides[0] == 8:
        pass
    else:
        raise RuntimeError("Design matrix is not C or F contiguous")
    normal_matrix=dsyrk(1.0,dadesign.data,trans=trans,lower=lower)

    #create the right hand side vector(s)
    rhs=dadesign.T@daobs

    #allow for multiple right hand sides
    ltpl=daobs.dot(daobs,dim=obsdim)
    
    #set number of observations and unknown parameters (ensure integer type)
    nobs=daobs.sizes[fwdop.obs_dim]*np.ones(nrhs,dtype=np.int64)
    npara=normal_matrix.shape[0]*np.ones(nrhs,dtype=np.int64)
    
    #tag auxiliary coords in rhs as xinv_types
    if len(rhs.shape) > 1: 
        auxdim=[dim for dim in rhs.dims if dim != unkdim][0]
        #add auxialiary_coord attributes to possible coordinate
        rhs.coords[auxdim].attrs.update(aux_coords_attrs(state='linked'))
    
    xinvcoords=find_xinv_coords(dsdesign)
    #update with the relevant coordinates from the righthand side matrix
    xinvcoords.update({name:coord for name,coord in rhs.coords.items() if name not in xinvcoords.keys()})
    dsout=xr.Dataset(dict(N=((unkdim,unkdim+"_"),normal_matrix.data),rhs=(rhs.dims,rhs.data),ltpl=((nrhsdim),ltpl.data),sigma0=((nrhsdim),sigma0.data),nobs=((nrhsdim),nobs.data),npara=((nrhsdim),npara.data)),coords=xinvcoords)
    #for some reason the coordinate attributes do not get properly propagated
    #so make sure they are added
    for key,coord in xinvcoords.items():
        dsout[key].attrs.update(coord.attrs)
    
    #create an apriori solution (zero)
    dsout['x0']=xr.zeros_like(dsout.rhs.reset_index(unkdim))
    dsout.x0.attrs.update(x0_attrs())

    #add attributes
    dsout.N.attrs.update(N_attrs(lower))
    dsout.rhs.attrs.update(rhs_attrs())
    dsout.ltpl.attrs.update(ltpl_attrs('apriori'))
    dsout.sigma0.attrs.update(sigma0_attrs('apriori'))
    dsout.nobs.attrs.update(nobs_attrs())
    dsout.npara.attrs.update(npara_attrs())
    return dsout




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import numpy as np
import xarray as xr
from xinv.core.attrs import find_xinv_unk_coord,get_state,get_type,xinv_tp,xinv_st,find_xinv_group_coords
from xinv.core.logging import xinvlogger

def find_ilocs(dsneq,dim,elements):
    idxname=f"_{dim}_idx"
    if idxname not in dsneq.keys():
        dsneq[idxname]=(dim,np.arange(dsneq.sizes[dim]))
    return dsneq[idxname].loc[elements].data

def find_overlap_coords(coord1,coord2):
    """
        Find the unique and common coordinates between two xarray coordinates
    """
    unkdim1=coord1.dims[0]
    unkdim2=coord2.dims[0]

    if unkdim1 != unkdim2:
        raise ValueError("Coordinates must have the same dimension name")
    if coord1.dtype != coord2.dtype:
        raise ValueError(f"Coordinates must have the same types, supplied are coord1: {coord1.dtype},coord2: {coord2.dtype}")

    intersect=np.intersect1d(coord1.data,coord2.data,return_indices=False)
    uniq1=np.setdiff1d(coord1, coord2, assume_unique=False)
    uniq2=np.setdiff1d(coord2, coord1, assume_unique=False)

    return uniq1, intersect, uniq2


def find_unk_idx(dsneq,sort=True,**kwargs):
    """ 
        Find the indices of a set of unknown parameters in the unknown vector of a normal equation system 
        Parameters:
        -----------
        dsneq: xarray.Dataset
            Dataset containing the normal equation system
        sort: bool, optional
            If True, the output indices are sorted in ascending order. The default is True.
        kwargs: dict
            keyword arguments with dimension names as keys and the elements to find as values
            e.g. poly=[0,1],harmonics_seasonal=[1,2]
        Returns:
        --------
        idxfound: np.ndarray or None
            Indices of the found unknown parameters in the unknown vector, or None if no parameters were found
        idxremaining: np.ndarray or None
            
            Indices of the unknown parameters that are complementary to the ones found
        idxnotfound: np.ndarray or None
            Indices of the requested unknown parameters that were not found in the system
            
    """

    xunk_co=find_xinv_unk_coord(dsneq)
    
    unkdim=xunk_co.dims[0]
    
    group_id_co=None
    group_seq_co=None

    notfound=[]
    found=[]
    remaining=[]
    for coname,searchparams in kwargs.items():
        co_search=dsneq[coname]
        if get_type(co_search) != xinv_tp.unk_co:
            raise ValueError(f"Missing xinv_type: supplied coordinate name {coname} has no valid relation with unknown coordinate {xunk_co.name}")
        dimname=co_search.dims[0]
        if searchparams is not type(xr.DataArray):
            #turn into DataArray
            searchparams=xr.DataArray(searchparams,dims=dimname)
        #find unique and overlapping coordinates over the unknown dimension
        notfnd,fnd,remng=find_overlap_coords(searchparams,co_search)
       
        if get_state(co_search) == xinv_st.unlinked:
            #we may have to apply an additional lookup in the group unknown multiindex
            if group_id_co is None and group_seq_co is None:
                group_id_co,group_seq_co=find_xinv_group_coords(dsneq)
            notfound.extend([(coname,i) for i in find_ilocs(dsneq,coname,notfnd)])
            found.extend([(coname,i) for i in find_ilocs(dsneq,coname,fnd)])
            remaining.extend([(coname,i) for i in find_ilocs(dsneq,coname,remng)])

        elif get_state(co_search) == xinv_st.linked:
            notfound.extend(notfnd)
            found.extend(fnd)
            remaining.extend(remng)
        else:
            raise ValueError(f"Reduction coordinate {coname} has no valid link state")
    
    

    #index vector of the found parameters
    idxfound=find_ilocs(dsneq,unkdim,found) if len(found) > 0 else None
    idxnotfound=find_ilocs(dsneq,unkdim,notfound) if len(notfound) > 0 else None
    idxremaining=find_ilocs(dsneq,unkdim,remaining) if len(remaining) > 0 else None
    if sort:
        if idxfound is not None:
            idxfound=np.sort(idxfound)
        if idxremaining is not None:
            idxremaining=np.sort(idxremaining)
        if idxnotfound is not None:
            idxnotfound=np.sort(idxnotfound)
    
    return idxfound,idxremaining,idxnotfound



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

# This file is part of the shxarray software which is licensed
# under the Apache License version 2.0 (see the LICENSE file in the main repository)
# Copyright Roelof Rietbroek (r.rietbroek@utwente.nl), 2023
#


import logging
# xinv wide logger
xinvlogger=logging.getLogger("xinv")

ch = logging.StreamHandler()

# create formatter
formatter = logging.Formatter('%(name)s-%(levelname)s: %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
xinvlogger.addHandler(ch)


def debugging():
    return xinvlogger.getEffectiveLevel() == logging.DEBUG

def setInfoLevel():
    """Set logging level for both python and c++ to INFO severity"""
    xinvlogger.setLevel(logging.INFO)

def setDebugLevel():
    """Set logging level for both python and c++ to DEBUG severity"""
    xinvlogger.setLevel(logging.DEBUG)


def setWarningLevel():
    """Set logging level for both python and c++ to WARNING severity"""
    xinvlogger.setLevel(logging.WARNING)

def setErrorLevel():
    """Set logging level for both python and c++ to WARNING severity"""
    xinvlogger.setLevel(logging.ERROR)

setInfoLevel()



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl, 2025 Kiana Karimi, s.karimi@utwente.nl 

import numpy as np
import xarray as xr
from xinv.fwd import FwdOpbase

class Harmonics(FwdOpbase):
    def __init__(self,freqs,harm_x="time",unknown_dim="harmonic",x0=None,delta_x=1,obs_dim=None,**kwargs):
        """ Setup a forward operator to represent multiple harmonic cosine and sine components."""
        if obs_dim is None:
            obs_dim=harm_x
        super().__init__(obs_dim=obs_dim, unknown_dim=unknown_dim,**kwargs)
        self._freqs=freqs
        self._x0=x0
        self._delta_x=delta_x
        self._n=2*len(self._freqs)
        self._harm_x=harm_x

    def _jacobian_impl(self,**kwargs):
        """ Creates the Jacobian of the forward operator (note: linear operator)."""
        
        #figure out xcoords

        #figure out the xcoords to use for the polynomial
        if self._harm_x in kwargs:
            #xcoords is directly supplied
            xcoords=kwargs[self._harm_x]
        elif "daobs" in kwargs:
            xcoords=kwargs['daobs'].coords[self._harm_x]
        else:
            raise ValueError(f"Harmonic Jacobian operator cannot figure out xcoord values, provide either dataarray 'daobs' or {self._harm_x} coordinate")
        order='C'
        jacobian=xr.DataArray(np.zeros([len(xcoords),self._n], order=order), dims=[self._obsdim,self._unkdim], name="harmonic_jacobian", coords={self._obsdim:xcoords,self._unkdim:np.arange(self._n)})
        
        if self._x0 is None:
            self._x0=np.mean(xcoords).item()          
            
        for i, freq in enumerate(self._freqs):
            omega_t=freq*(((xcoords-self._x0)/self._delta_x).astype(np.float64))
            jacobian.loc[:,2*i]=np.cos(omega_t)
            jacobian.loc[:,2*i+1]=np.sin(omega_t)
            
        return jacobian
    
class SeasonalHarmonics(Harmonics):
    def __init__(self,x0,semi_annual=True,harm_x="time",delta_x=np.timedelta64(365,'D')+np.timedelta64(int(86400/4),'s')):
        
        freqs=[2*np.pi]
        if semi_annual is True:
            freqs.append(4*np.pi)
            
        super().__init__(freqs=freqs,unknown_dim="harmonics_seasonal",harm_x=harm_x,delta_x=delta_x,x0=x0)

        


import pytest
import xarray as xr
import numpy as np
from xinv import *
import os
from xinv.core.tools import find_ilocs
from xinv.core.attrs import find_xinv_coords,xinv_tp,xinv_st,find_neq_components
from xinv.neq import transform
from xinv.fwd.matrixfwd import MatrixfwdOp
import pandas as pd
from xinv.core.exceptions import XinvIllposedError
from xinv.fwd.polynomial import Polynomial

from fixtures import noisypoly


def test_neq_transform(noisypoly):
    """
    Test the NEQ transform where all parameters are transformed
    """

    #build a base neq system with a polynomial of degree 0 (just a mean)
    x0=noisypoly.attrs['x0']
    deltax=noisypoly.attrs['delta_x']
    npolybase=0
    unkdim='polybase'
    polyfwdbase=Polynomial(n=npolybase,poly_x='x',x0=x0,delta_x=deltax,unknown_dim=unkdim)
    
    #amount of polynomials we would like to resolve after transformation
    npoly_resolve=noisypoly.sizes['poly']-1

    #create a transformation which relates the mean to a polynomial of npoly
    
    std_noise=0.5
    #polynomial forward operator to use in the transformation, this maps a set of higher degree polynomial coefficients (new unknowns) to a mean (previous uknown)
    polyfwd=Polynomial(n=npoly_resolve,poly_x='x',x0=x0,delta_x=deltax,cache=False,obs_dim=unkdim,auxcoords={unkdim:(unkdim,np.arange(npolybase+1))}) #note cache is false because it needs to be recomputed everytime the jacobian is constructed

    neqtrans=None

    # polyfwd2=Polynomial(n=npoly_resolve,poly_x='x',x0=x0,delta_x=deltax,cache=True)
    # dsneq=noisypoly.polyobs.xi.build_normal(polyfwd2,ecov=std_noise*std_noise)
    #solve the joint system
    # dssolcomb=dsneq.xi.solve()
    
    #loop over x slices (make one neq system per x slice and transform and add it)
    for x in noisypoly.x.values:
        daobs=noisypoly.polyobs.sel(x=[x])
        neqsingle=daobs.xi.build_normal(polyfwdbase,ecov=std_noise*std_noise) 
        #transform
        if neqtrans is None:
            neqtrans=neqsingle.xi.transform(polyfwd,x=[x])
        else:
            #add all contributions
            neqtrans=neqtrans.xi.add(neqsingle.xi.transform(polyfwd,x=[x]))
    
    dssol=neqtrans.xi.solve()
    
    #compare with combined system (should give the same results)
    polyfwd2=Polynomial(n=npoly_resolve,poly_x='x',x0=x0,delta_x=deltax,cache=True)
    dsneq=noisypoly.polyobs.xi.build_normal(polyfwd2,ecov=std_noise*std_noise)
    
    #solve the joint system
    dssolcomb=dsneq.xi.solve()
    prenoise=noisypoly.attrs['noise_std']
    tol=3*np.sqrt(np.diag(dssol.COV))*prenoise
    tol=tol.max()
    assert np.allclose(dssol.solution,noisypoly.polytrue,atol=tol)        

    #also check covariance
    assert np.allclose(dssol.COV,dssolcomb.COV)        

def test_neq_partial_transform(noisypoly):
    """
    Test the NEQ transform where some parameters are transformed
    """

    #build a base neq system with a polynomial of degree 0 (just a mean)
    x0=noisypoly.attrs['x0']
    deltax=noisypoly.attrs['delta_x']
    npolybase=0
    unkdim='polybase'
    polyfwdbase=Polynomial(n=npolybase+1,poly_x='x',x0=x0,delta_x=deltax,unknown_dim=unkdim)
    
    #amount of polynomials we would like to resolve after transformation
    npoly_resolve=noisypoly.sizes['poly']-1

    #create a transformation which relates the mean to a polynomial of npoly
    
    std_noise=0.5
    #polynomial forward operator to use in the transformation, this maps a set of higher degree polynomial coefficients (new unknowns) to a mean (previous uknown)
    polyfwd=Polynomial(n=npoly_resolve,poly_x='x',x0=x0,delta_x=deltax,cache=False,obs_dim=unkdim,auxcoords={unkdim:(unkdim,np.arange(npolybase+1))}) #note cache is false because it needs to be recomputed everytime the jacobian is constructed

    neqtrans=None

    
    #loop over x slices (make one neq system per x slice and transform it)
    for x in noisypoly.x.values:
        daobs=noisypoly.polyobs.sel(x=[x])
        neqsingle=daobs.xi.build_normal(polyfwdbase,ecov=std_noise*std_noise) 
        
        #transform
        if neqtrans is None:
            neqtrans=neqsingle.xi.transform(polyfwd,x=[x])
        else:
            #add all contributions
            neqtrans=neqtrans.xi.add(neqsingle.xi.transform(polyfwd,x=[x]))
    
    
    #try solving (should not work because of rank defect
    try:
        dssol=neqtrans.xi.solve()
        #should not get here:
        assert False
    except XinvIllposedError: 
        #expected behavior
        pass


    #select only the new 'poly' group (which should be solvable)
    neqtranspoly=neqtrans.xi.get_group('poly')

    dssol=neqtranspoly.xi.solve()
   
    #compare with combined system (should give the same results)
    polyfwd2=Polynomial(n=npoly_resolve,poly_x='x',x0=x0,delta_x=deltax,cache=True)
    dsneq=noisypoly.polyobs.xi.build_normal(polyfwd2,ecov=std_noise*std_noise)
    #solve the joint system
    dssolcomb=dsneq.xi.solve()
    
    prenoise=noisypoly.attrs['noise_std']
    tol=3*np.sqrt(np.diag(dssol.COV))*prenoise
    tol=tol.max()
    assert np.allclose(dssol.solution,noisypoly.polytrue,atol=tol)        
    
    #also check covariance
    assert np.allclose(dssol.COV,dssolcomb.COV)        



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, rietbroek@utwente.nl


from xinv.core.attrs import ltpl_attrs,sigma0_attrs,N_attrs,rhs_attrs,nobs_attrs,npara_attrs,xinv_tp,x0_attrs
from xinv.core.logging import xinvlogger

import numpy as np
import xarray as xr

def zeros(rhsdims,coords,lower=0,norder='C'):
        
        
        #figure out the shape of the rhs from the provided coordinates
        # rhsshape=[len(coords[dim]) for dim in rhsdims]
        rhsshape=[]
        grpco=None
        for i,dim in enumerate(rhsdims):
            if dim not in coords:
                raise KeyError(f"Coordinate {dim} not found in the provided coordinates")
            if type(coords[dim]) == xr.core.coordinates.Coordinates:
                #Extract correct subcoordinate from the multindex group coordinate
                co=coords[dim][dim]
                grpco=coords[dim]
            else:
                co=coords[dim]
            #Note: We assume that the first dimension is the unknown parameter dimension
            if i== 0 and not "xinv_type" in co.attrs:
                xinvlogger.warning(f"Unknown parameter dimension is assumed to be the first one {rhsdims[0]}")
            elif i==0 and co.attrs["xinv_type"] != xinv_tp.unk_co:
                raise ValueError("The unknown parameter does not correspond to the first dimenion of the right hand side")
            rhsshape.append(len(co))


        #allocate space
        N=([rhsdims[0],rhsdims[0]+'_'],np.zeros([rhsshape[0],rhsshape[0]],order=norder))
        rhs=(rhsdims,np.zeros(rhsshape))
        x0=(rhsdims,np.zeros(rhsshape))

        ltpl=(rhsdims[1:],np.zeros(rhsshape[1:]))
        sigma0=(rhsdims[1:],np.zeros(rhsshape[1:]))
        nobs=(rhsdims[1:],np.zeros(rhsshape[1:],dtype=np.int64))
        npara=(rhsdims[1:],np.zeros(rhsshape[1:],dtype=np.int64))
        
        if grpco is not None:
            dsneq=xr.Dataset(dict(N=N,rhs=rhs,x0=x0,ltpl=ltpl,nobs=nobs,npara=npara,sigma0=sigma0),coords=grpco).assign_coords({ky:co for ky,co in coords.items() if ky != rhsdims[0]})
        else:
            dsneq=xr.Dataset(dict(N=N,rhs=rhs,x0=x0,ltpl=ltpl,nobs=nobs,npara=npara,sigma0=sigma0),coords=coords)
        #add attributes
        dsneq.N.attrs.update(N_attrs(lower))
        dsneq.rhs.attrs.update(rhs_attrs())
        dsneq.x0.attrs.update(x0_attrs())
        dsneq.ltpl.attrs.update(ltpl_attrs('apriori'))
        dsneq.sigma0.attrs.update(sigma0_attrs('apriori'))
        dsneq.nobs.attrs.update(nobs_attrs())
        dsneq.npara.attrs.update(npara_attrs())

        return dsneq


    


    
    



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import xarray as xr
import numpy as np
from xinv.fwd import FwdOpbase

class Polynomial(FwdOpbase):
    def __init__(self,n,poly_x='x',x0=None,delta_x=1,cache=False,unknown_dim="poly",obs_dim=None,**kwargs):
        """Setup a forward operator which represents a polynomial of degree n"""
        if obs_dim is None:
            obs_dim=poly_x
        super().__init__(obs_dim=obs_dim,unknown_dim=unknown_dim,cache=cache,**kwargs)
        self._n=n
        self._unkcoord=np.arange(self._n+1)
        self._delta_x=delta_x
        self._x0=x0
        self._poly_x=poly_x

    def _jacobian_impl(self,daobs=None,x=None, auxcoords=None):
        """Creates the Jacobian of the forward operator (note:linear operator)"""
        #figure out the xcoords to use for the polynomial
        if x is not None:
            xcoords=np.asarray(x)
        elif daobs is not None and self._poly_x in daobs.coords:
            xcoords=daobs.coords[self._poly_x].data
        else:
            raise ValueError(f"Polynomial Jacobian operator cannot figure out xcoord values, provide either dataarray 'daobs' or through a xcoords argument to the Jacobian")

        if self._x0 is None:
            self._x0=np.mean(xcoords).item()
       
        #possibly copy auxiliary coordinates from the input
        if auxcoords is not None:
            coords={ky:val for ky,val in auxcoords.items()}
        else:
            coords={}
        #make sure that the xcoords share the obsdim
        coords[self._poly_x]=(self._obsdim,xcoords)
        coords[self._unkdim]=(self._unkdim,self._unkcoord)
        

        order='C'
        jacobian=xr.DataArray(np.zeros([len(xcoords),self._n+1],order=order),dims=[self._obsdim,self._unkdim],name="poly_jacobian",coords=coords)
        #normalize xcoords to proposed time scale
        xcoords_rel=((xcoords-self._x0)/self._delta_x).data
        #fill polynomial scales
        for i in range(self._n+1):
            jacobian.loc[:,i]=np.power(xcoords_rel,i)
        return jacobian






# This file is part of the shxarray software which is licensed
# under the Apache License version 2.0 (see the LICENSE file in the main repository)
# Copyright Roelof Rietbroek (r.rietbroek@utwente.nl), 2023
#




from setuptools import setup,Extension
from setuptools_scm import get_version
from Cython.Build import cythonize
import Cython.Compiler.Options
from packaging.version import Version

import os 
import numpy as np
import sys

if sys.platform.startswith("win"):
    winplatform=True
    extra_args = ['/openmp']
else:
    winplatform=False
    extra_args = ['-fopenmp']

if "DEBUG_CYTHON" in os.environ:
    debug=True
    extra_args.append('-O0')
    #extra_args.append('-pg')
else:
    debug=False
    extra_args.append('-O3')

useCython=True
if useCython:
    ext='.pyx'
else:
    ext=".cpp"

def listexts():
    names=["xiext"]
    exts=[]
    for nm in names:
        exts.append(Extension("xinv."+nm.replace("/","."),["src/extension/"+nm+ext],include_dirs=[np.get_include(),"."], define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],extra_compile_args=extra_args,extra_link_args=extra_args))
    return exts

extensions=listexts()


if useCython:
    #additionally cythonize pyx files before building
    extensions=cythonize(extensions,language_level=3,annotate=True,gdb_debug=debug,compiler_directives={'embedsignature': True})

setup(
    version = get_version(root='.', relative_to=__file__),
    ext_modules=extensions
    )



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl



from xarray.backends import BackendEntrypoint
from xinv.io.sinex import read_sinex

class SINEXBackEntryPoint(BackendEntrypoint):
    url="https://github.com/strawpants/xinv"
    description = "Read normal equation systems/solutions in SINEX (2.02) format"
    def open_dataset(self,filename_or_obj,*,drop_variables=None):
        if drop_variables is not None and "N" in drop_variables:
            #Special case: it is much quicker to abandon reading when a matrix is present and not needed
            dsout=read_sinex(filename_or_obj,stopatmat=True)
            drop_variables
        elif drop_variables is not None:
            dsout=read_sinex(filename_or_obj)
            dsout=dsout.drop_vars(drop_variables)
        else:
            dsout=read_sinex(filename_or_obj)
        
        return dsout
    
    def guess_can_open(self,filename_or_obj):
        try:
            strrep=str(filename_or_obj).lower()
            # search for usual file naming of SINEX files
            if strrep.endswith('.snx') or strrep.endswith('.snx.gz') or strrep.endswith('.snx.z'):  
                #Found a file name which probably is a sinex file
                return True
        except AttributeError:
            return False
            
        return False




## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import xarray as xr
from xinv.core.attrs import islower,Chol_attrs,xinv_st,cov_attrs
from xinv.core.logging import xinvlogger
import numpy as np
from xinv.core.exceptions import XinvIllposedError

from scipy.linalg import cholesky
from scipy.linalg.lapack import dpotri
from scipy.linalg.blas import dtrsm,dsyrk,dsymm

import ctypes
from enum import Enum

class MemLayout(Enum):
    F_cont = 1
    C_cont = 0
    non_cont = -1

def memlayout(arr:np.ndarray):
    if arr.flags['F_CONTIGUOUS'] and arr.strides[0] == arr.dtype.itemsize:
        return MemLayout.F_cont
    elif arr.flags['C_CONTIGUOUS'] and arr.strides[1] == arr.dtype.itemsize:
        return MemLayout.C_cont
    else:
        return MemLayout.non_cont


def cholesky_inplace(N:xr.DataArray):
    """
        Inplace Cholesky decomposition of a symmetric positive definite matrix stored in an xarray DataArray. 
        The input matrix N is overwritten with its Cholesky factor.

    """


    if N.attrs['xinv_state'] not in [xinv_st.symU,xinv_st.symL]:
        raise RuntimeError(f"Don't know how to do a Cholesky decomposition on this xinv state :{N.attrs['xinv_state']}")
   
    lower=islower(N)
    restore=False

    n_layout=memlayout(N.data)

    #check for C or F ordering
    if n_layout == MemLayout.F_cont:
        #fine F ordered array with stride 1 of double values
        lowapparent=lower
        #pointer
        Ndat=N.data
    elif n_layout == MemLayout.C_cont:
        #switch lower/upper representation in the F-ordered matrix of doubl
        lowapparent=1-lower
        Ndat=N.data.T
    else:
        xinvlogger.warning("Normal matrix is not C or F contiguous, applying copy-restore")
        lowapparent=lower
        restore=True
        Ndat=N.data.copy(order='F')

    try:
        cholesky(Ndat,lower=lowapparent,overwrite_a=1)
        if restore:
            #copy data back
            N[()]=Ndat
    except np.linalg.LinAlgError as e:
        raise XinvIllposedError(str(e))
    N.attrs.update(Chol_attrs(lower))

    return N #same object (will be changed now)

def dpotri_inplace(N:xr.DataArray):
    lower=islower(N)

    restore=False

    n_layout=memlayout(N.data)
    #check for C or F ordering
    if n_layout == MemLayout.F_cont:
        #fine F ordered array with stride 1 of double values
        lowapparent=lower
        #pointer
        Ndat=N.data
    elif n_layout == MemLayout.C_cont:
        #switch lower/upper representation in the F-ordered matrix of doubl
        lowapparent=1-lower
        Ndat=N.data.T
    else:
        xinvlogger.warning("Normal matrix is not C or F contiguous, applying copy-restore")
        lowapparent=lower
        restore=True
        Ndat=N.data.copy(order='F')
    dpotri(Ndat,lower=lowapparent,overwrite_c=1)
    if restore:
        #copy data back
        N[()]=Ndat 
    N.attrs.update(cov_attrs()) 
    return N




def dtrsm_inplace(Chol:xr.DataArray,rhs:xr.DataArray,trans=0):
    """
    Triangular solve a RHS matrix with the Cholesky factor 
    """
    restore=False
    rhs_layout=memlayout(rhs.data)
    
    
    if rhs_layout == MemLayout.F_cont:
        rhsdat=rhs.data
        transa=trans
        side=0 #0 means left
    elif rhs_layout == MemLayout.C_cont:
        rhsdat=rhs.data.T
        transa=1-trans
        side=1 #1 means right  
    else:
        xinvlogger.warning("Right hand side matrix is not C or F contiguous, applying copy and restore")
        rhsdat=rhs.data.copy(order='F')
        transa=0
        side=1 #0 means left  
        restore=True

    lower=islower(Chol)
    chol_layout=memlayout(Chol.data)
    if chol_layout == MemLayout.C_cont:
        #To prevent an additional copy by f2py we can fake a F contigous array by transposing and switching lower/upper
        lower=1-lower
        transa=1-transa
        choldat=Chol.data.T
    else:
        choldat=Chol.data
    
    dtrsm(1.0,choldat,rhsdat,side=side,lower=lower,trans_a=transa,overwrite_b=1)

    if restore:    
        rhs.data[()]=rhsdat

    return rhs #although changed in place

def dsyrk_inplace(N,A,trans=0,beta=0.0,alpha=1.0):
    
    restore=False
    lower=islower(N)
    
    n_layout=memlayout(N.data)

    #check for C or F ordering
    if n_layout == MemLayout.F_cont:
        #fine F ordered array with stride 1 of double values
        #pointer
        Ndat=N.data
    elif n_layout == MemLayout.C_cont:
        #switch lower/upper representation in the F-ordered matrix of doubl
        lower=1-lower
        Ndat=N.data.T
    else:
        xinvlogger.warning("Normal matrix is not C or F contiguous, applying copy-restore")
        restore=True
        Ndat=N.data.copy(order='F')
    
    a_layout=memlayout(A.data)
    
    
    if a_layout == MemLayout.F_cont:
        adat=A.data
        transa=trans
    elif a_layout == MemLayout.C_cont:
        adat=A.data.T
        transa=1-trans
        side=1 #1 means right  
    else:
        xinvlogger.warning("Right hand side matrix is not C or F contiguous, applying copy and restore")
        adat=A.data.copy(order='F')
        transa=0
        restore=True
    dsyrk(alpha=alpha, a=adat,beta=beta, c=Ndat, trans=transa, lower=lower, overwrite_c=1)
    
    if restore:    
        A[()]=adat

    return N

def dsymm_inplace(A:xr.DataArray,B:xr.DataArray,C:xr.DataArray,alpha=1.0,beta=0.0):
    """
    Symmetric matrix times matrix multiplication
        C= alpha*A*B +beta*C
    """
    #
    #    C := alpha*A*B + beta*C, (side=0)
    #or
    #    C := alpha*B*A + beta*C, (side=1)
    
    #some basic checks
    if A.shape[0] != A.shape[1]:
        raise ValueError("Matrix A must be square")
    if A.shape[1] != B.shape[0]:
        raise ValueError("Inner dimensions of A and B do not match")
    if A.shape[0] != C.shape[0]:
        raise ValueError("Output matrix C rows do not match matrix A")
    
    if B.ndim != C.ndim:
        raise ValueError("B and C must have the same number of dimensions")

    for i in range(B.ndim-1):
        if B.shape[i] != C.shape[i]:
            raise ValueError(f"Dimensions of C  do not match that of B at axis {i}")
    # if B.ndim == 2 and C.ndim == 2:
        # if B.shape[1] != C.shape[1]:
            # raise ValueError("Columns of C  do not match that of B")
    # elif B.ndim ==1 and C.ndim ==1:
        # if B.shape[0] != C.shape[0]:
            # raise ValueError("Dimensions of C  do not match that of B")
    

    restore=False
    c_layout=memlayout(C.data)
    
    if c_layout == MemLayout.F_cont:
        cdat=C.data
        side=0 #0 means left
    elif c_layout == MemLayout.C_cont:
        cdat=C.data.T
        #switch side to allow for in place operation
        side=1
    else:
        xinvlogger.warning("C matrix is not C or F contiguous, applying copy and restore")
        cdat=C.data.copy(order='F')
        side=0 #0 means left  
        restore=True

    lower=islower(A)
    A_layout=memlayout(A.data)
    if A_layout == MemLayout.C_cont:
        #To prevent an additional copy by f2py we can fake a F contigous array by transposing and switching lower/upper
        lower=1-lower
        adat=A.data.T
    else:
        adat=A.data
    
    
    if side == 1:
        bdat=B.data.T
    else:
        bdat=B.data
    
    b_layout=memlayout(B.data)
    

    if b_layout != c_layout:
        xinvlogger.warning("B matrix contiguousness is inconsistent with C, a copy will be made by f2py")
    #dsymm(alpha, a, b[, beta, c, side, lower, overwrite_c])
    dsymm(alpha,adat,bdat,beta=beta,c=cdat,side=side,lower=lower,overwrite_c=1)


    if restore:    
        C.data[()]=cdat

    return C #although changed in place



# Test some basic operations of xarray objects filled with spherical harmonic datasets
# This file is part of the shxarray software which is licensed
# under the Apache License version 2.0 (see the LICENSE file in the main repository)
# Copyright Roelof Rietbroek (r.rietbroek@utwente.nl), 2023


import pytest
import xarray as xr
import os
import numpy as np
import time
from xinv.core.logging import xinvlogger



@pytest.fixture
def sinexvaldata(request):
    if request.param == "ITSG":
        # some validation values based on prior visual inspection of sinex file
        #+SOLUTION/ESTIMATE
        # ...
        # 91 CN        9 --    7 03:259:00000 ---- 2 -1.17979570334447e-07 1.01491e-12
        # 92 SN        9 --    7 03:259:00000 ---- 2 -9.69270369824925e-08 1.02558e-12
        # ..
        #+SOLUTION/APRIORI
        #...
        # 25 CN        5 --    2 03:259:00000 ---- 2  6.52120740523400e-07 0.00000e+00
        # 26 SN        5 --    2 03:259:00000 ---- 2 -3.23349434999185e-07 0.00000e+00
        #..
        #+SOLUTION/NORMAL_EQUATION_VECTOR
        #*INDEX _TYPE_ CODE PT SOLN _REF_EPOCH__ UNIT S ___RIGHT_HAND_SIDE___
        #1 CN        2 --    0 03:259:00000 ---- 2 -3.27811136401904e+12
        #2 CN        2 --    1 03:259:00000 ---- 2 -4.32367855180938e+11
        #+SOLUTION/NORMAL_EQUATION_MATRIX U
        # ....
        # 7     7  2.92598076849103e+23 -3.55351048890917e+21 -3.78432268065340e+21
        # 7    10 -6.34992891236174e+21  2.34638506385098e+20  4.78761459861900e+21
        # ...
        valdata={} 
        valdata["sol_est"]=[(dict(n=9,m=7), -1.17979570334447e-07),(dict(n=9,m=-7),-9.69270369824925e-08)] 
        valdata["sol_std"]=[(dict(n=9,m=7),1.01491e-12),(dict(n=9,m=-7), 1.02558e-12)]
        valdata["apri_est"]=[(dict(n=5,m=2),6.52120740523400e-07),(dict(n=5,m=-2),-3.23349434999185e-07)] 
        valdata["rhs"]=[(dict(n=2,m=0),-3.27811136401904e+12),(dict(n=2,m=1),-4.32367855180938e+11)]
        valdata["N"]=[(7,7, 2.92598076849103e+23),(7,8,-3.55351048890917e+21),(7,9,-3.78432268065340e+21),(7,10,-6.34992891236174e+21),(7,11,2.34638506385098e+20),(7,12,4.78761459861900e+21)]
        
    #possibly also download the file and pass the file name

        url="https://ftp.tugraz.at/outgoing/ITSG/GRACE/ITSG-Grace2018/monthly/normals_SINEX/monthly_n96/ITSG-Grace2018_n96_2003-09.snx.gz"
    elif request.param == "IGS":
        # 36 STAZ   ALBH  A    3 22:351:43185 m    1 0.474579124249256E+07 .318348E-03
    # 37 STAX   ALGO  A    1 22:349:43185 m    1 0.918129084134536E+06 .184114E-03
    # 38 STAY   ALGO  A    1 22:349:43185 m    1 -.434607133252970E+07 .372889E-03
        valdata={} 
        valdata["sol_est"]=[(dict(statype="STAZ",site="ALBH",monument="A",soluid=3), 0.474579124249256E+07),
                            (dict(statype="STAX",site="ALGO",monument="A",soluid=1), 0.918129084134536E+06),
                            (dict(statype="STAY",site="ALGO",monument="A",soluid=1), -0.434607133252970E+07)]

        valdata["sol_std"]=[(dict(statype="STAZ",site="ALBH",monument="A",soluid=3), 0.318348E-03),
                            (dict(statype="STAX",site="ALGO",monument="A",soluid=1), 0.184114E-03),
                            (dict(statype="STAY",site="ALGO",monument="A",soluid=1), 0.372889E-03)]
    # 32 STAY   ALBH  A    2 22:350:43185 m    1 -.353904953666000E+07 .000000E+00
    # 33 STAZ   ALBH  A    2 22:350:43185 m    1 0.474579123972000E+07 .000000E+00
    # 34 STAX   ALBH  A    3 22:351:43185 m    1 -.234133313693707E+07 .000000E+00
        valdata["apri_est"]=[(dict(statype="STAY",site="ALBH",monument="A",soluid=2), -0.353904953666000E+07),
                            (dict(statype="STAZ",site="ALBH",monument="A",soluid=2), 0.474579123972000E+07),
                            (dict(statype="STAX",site="ALBH",monument="A",soluid=3), -0.234133313693707E+07)]
    # 2433 STAZ   ZIM3  A    3 22:351:43185 m    2 -.190963469354183E-01
    # 2434 STAX   ZIMM  A    1 22:351:43185 m    2 0.622005988863635E-02
    # 2435 STAY   ZIMM  A    1 22:351:43185 m    2 -.455479611760775E-01
        valdata["rhs"]=[(dict(statype="STAZ",site="ZIM3",monument="A",soluid=3), -0.190963469354183E-01),
                            (dict(statype="STAX",site="ZIMM",monument="A",soluid=1), 0.622005988863635E-02),
                            (dict(statype="STAY",site="ZIMM",monument="A",soluid=1), -0.455479611760775E-01)]
    # 46    40 -0.40477609814186E-02 -0.14917111114273E-02 -0.16167305540122E-02
    # 46    43 -0.40488191574758E-02 -0.54300924437564E-03  0.16845406308024E-02
    # 46    46  0.32043242340174E+02

        valdata["N"]=[(46,40, -0.40477609814186E-02),(46,41,-0.14917111114273E-02),(46,42,-0.16167305540122E-02),
                      (46,43, -0.40488191574758E-02),(46,44,-0.54300924437564E-03),(46,45, 0.16845406308024E-02),
                      (46,46,0.32043242340174E+02)]

        url="https://igs.bkg.bund.de/root_ftp/IGSac/products/2240/COD0OPSFIN_20223500000_01D_01D_SOL.SNX.gz"



    sinexfile=os.path.join(os.path.dirname(__file__),'testdata', os.path.basename(url))
    
    if not os.path.exists(sinexfile):
        import requests
        xinvlogger.info(f"Downloading {sinexfile}...")
        r=requests.get(url)
        with open(sinexfile,'wb') as fid:
            fid.write(r.content)
    return valdata,sinexfile


     

@pytest.mark.parametrize("sinexvaldata",["IGS","ITSG"],indirect=True)
def test_sinex(sinexvaldata):
    """ Test reading of Normal equation systems in SINEX format"""
    sinexval,sinexfile=sinexvaldata
    #Engine does not need to be specified because file corresponds to commonly used filename pattern for sinex 
    start=time.time()
    dsneqsinex=xr.open_dataset(sinexfile,drop_variables=["N"])
    end=time.time()
    xinvlogger.info(f"Time to read sinex file without normal matrix {end-start:.2f} seconds")

    if 'COD0OPSFI' in sinexfile:
        #get a subgrpup
        dsneqsinex=dsneqsinex.xi.get_group("stat")

    for var in ["sol_est","apri_est","sol_std","rhs"]:
        for midict,val in sinexval[var]:
            assert val == dsneqsinex[var].sel(midict).item()
        # assert dsneqsinex[var].sel(n=
    
    #read version with entire normal matrix
    start=time.time()
    dsneqsinex=xr.open_dataset(sinexfile)
    end=time.time()
    xinvlogger.info(f"Time to read sinex file with normal matrix {end-start:.2f} seconds")
    for ix,iy,val in sinexval["N"]:
        #note index ix,and iy are 1-indexed
        assert val == dsneqsinex.N[ix-1,iy-1].item()



